/*                               -*- Mode: C -*-
 * bt-structure.c -- Functions used for and with op-structure.
 *
 * Copyright (c) 2024-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#define _GNU_SOURCE

#include "slistPack.h"
#include "shashPack_f.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <err.h>

#include "version.h"
#include "constant.h"
#include "bt2f-sprint.h"
#include "bt-structure.h"

#include "bt-structure_f.h"
#include "bt2f-type_f.h"
#include "bt2f-print_f.h"
#include "bt2f-sprint_f.h"

Symbol new_btnode_name(void);
Symbol new_btnode_id(BTNode_Type type);
//unsigned int handle_action_arg(BTNode *btnode);

char *current_bt_file_name = NULL;

HTick_Type htick = ACTION;


Slist *bt_list = NULL;		/* list of Beahavior Tree, i.e. the Top Level one(s). */
Slist *btnode_list = NULL;	/* list all the BT Node found */
Slist *btidactarg_list = NULL;


int bt_index = 2; 		/* First available node tcldot index */

/* the following code is a hack to synthesize the array of BT Type names (char *) */
#define T(x) #x,
const char * const BTNode_Type_Name[] = { BTTYPES };
#undef T

#define T(x) #x,
const char * const BTReturn_Type_Name[] = { BTRET };
#undef T

Slist *btn_stack;		/* Used while loading to change the proper btnode while doing recursive parsing. */

KeyVal *make_key_val(Symbol key, Term *val)
{
     KeyVal *res = MAKE_OBJECT(KeyVal);

     res->key = key;
     res->val = val;

     return res;
}

BTNode* current_btnode()
{
  return (BTNode* )sl_get_slist_head(btn_stack);
}


BTId_Action_Arg *make_btia(Symbol id)
{
  BTId_Action_Arg *t = MAKE_OBJECT(BTId_Action_Arg);

  t->id = id;
  t->arg_string =  NULL;

  return t;
}


BTNode *make_btnode()
{
  return make_btnode_type(UnknownBTNodeType);
}

BTNode *make_btnode_type(BTNode_Type type)
{
  BTNode *t = MAKE_OBJECT(BTNode);

  t->type = type;
  t->id = new_btnode_id(t->type); /* we set it, may be it will be overriden, no problem */
  t->name = new_btnode_name(); /* we set it, may be it will be overriden, no problem */
  t->id = new_btnode_id(t->type); /* we set it, may be it will be overriden, no problem */
  t->PossibleRet = SFR;
  t->children =  sl_make_slist();
  t->args =  sl_make_slist();
  t->pass_arg =  FALSE;
  t->halt =  FALSE;
  t->wait =  FALSE;
  t->tcldot = bt_index++;
  t->expr = NULL;
  t->u.needed_success = 0;

  return t;
}

void set_needed_success_if_not_set(BTNode *n)
{
  if (n->u.needed_success == 0) n->u.needed_success = sl_slist_length(n->children);
}


PBoolean equal_name_btnode_list (Symbol name, BTNode *n)
{
     return (name == n->name);
}

PBoolean equal_id_btnode_list (Symbol id, BTNode *n)
{
     return (id == n->id);
}

BTNode *find_btnode_or_create(Symbol name)
{
  BTNode *n;

  if ((n = (BTNode *)sl_search_slist(btnode_list,name,equal_name_btnode_list)) == NULL) {
     n = make_btnode();
     n->name = name;
  }
  return n;
}

Symbol btnode_name(BTNode *btnode)
{
	return btnode->name;	
}

PString btnode_file_name(BTNode *btnode)
{
	return btnode->file_name;	
}

/*
 * Free Functions
 */

BTNode *dup_btnode(BTNode *btnode)
{
     DUP(btnode);
     return btnode;
}

void free_btnode(BTNode *btnode)
{
  if (btnode) {
    if (LAST_REF(btnode)) {

      sl_loop_through_slist(btnode->children, btnode, BTNode *) {
	free_btnode(btnode);
      }

      FREE_SLIST(btnode->children);

      FREE(btnode->file_name);
    }

    FREE(btnode);
  }
}

static int node_name_index = 0;

Symbol new_btnode_name(void)
{
     char name[64];
     Symbol btnode_name;
     PBoolean save_cs = check_symbol;

     check_symbol = FALSE;

     do {
	  sprintf(name, "btn%d", node_name_index++);
	  DECLARE_ID(name, btnode_name);
     } while (sl_search_slist(btnode_list, btnode_name, equal_name_btnode_list) != NULL) ;

     check_symbol = save_cs;

     return btnode_name;
}

static int node_id_index = 0;

Symbol new_btnode_id(BTNode_Type type)
{
  char name[64];
  Symbol btnode_id;
  PBoolean save_cs = check_symbol;

  check_symbol = FALSE;

  const char *type_name = BTNode_Type_Name[type];

  do {
    sprintf(name, "%s%d", type_name, node_id_index++);
    DECLARE_ID(name, btnode_id);
  } while (sl_search_slist(btnode_list, btnode_id, equal_id_btnode_list) != NULL) ;

  check_symbol = save_cs;

  return btnode_id;
}


PBoolean sort_btnode(BTNode *btnode1, BTNode *btnode2)
{
     return (strcmp(btnode1->name,btnode2->name) < 0);
}

Event *current_event;
Sv *current_sv;
Slist *sv_list;
Res *current_res;
/* Skill *current_skill; */
/* Invariant *current_invariant; */
/* Failure *current_failure; */
/* Success *current_success; */
/* SkillSet *current_skillset; */

int index_level[10];		/* more than 10 level deep? unlikely... */

long ns_per_tick = 100000000;	/* should be set by the parser... */

long ns_per_second = 1000000000;

#define TICK_PER_RATE(rate) ((int)(ns_per_second/(ns_per_tick*(rate))))

#define TIME_IN_TICK(t) ((int)(lround(((t)*1000000000.)/(double)ns_per_tick)))

Transition *make_transition(State from, State to)
{
  Transition *t = MAKE_OBJECT(Transition);

  t->from = from;
  t->to = to;

  return t;
}

Event *make_event(Symbol name)
{
     Event *e = MAKE_OBJECT(Event);

     e->name = name;
     e->guard = (Expression *)NULL;
     e->effects = NULL;

     current_event = e;
     return e;
}

Sv *make_sv(Symbol name)
{

  Sv *f = MAKE_OBJECT(Sv);

  //  fprintf(stderr,"******** Making sv %s.\n", name);

     f->name = name;
     f->states = sl_make_slist();
     f->transitions = sl_make_slist();
     f->init = NULL;
     f->all_transitions = FALSE;

     f->min = -1;
     f->initi = -1;
     f->max = -1;

     current_sv = f;
     return f;
}

Res *make_res(Symbol name)
{
     Res *res = MAKE_OBJECT(Res);

     res->name = name;
     res->init = 0;
     res->max = 0;

     current_res = res;
     return res;
}


PBoolean valid_transition(State from, State to, Slist *transitions)
{
  Transition *t;

  sl_loop_through_slist(transitions, t, Transition *) {
    if (t->from == from && t->to == to) return TRUE;
  }
  return FALSE;
}


Time_Interval *make_time_interval(int min, int max, PBoolean max_inf)
{
     Time_Interval *ti = MAKE_OBJECT(Time_Interval);
     ti->min = min;
     ti->max = max;
     ti->max_inf = max_inf;

     return ti;
}

unsigned int action_arg_index = 1;

void add_event_to_event_list(void)
{
  /* Slist *event_list = current_skillset->event_list; */

  /* sl_add_to_tail(event_list, current_event); */
}


void add_sv_to_sv_list(void)
{

  if (!(current_sv->initi != -1 &&
       current_sv->min != -1 &&
       current_sv->max != -1)
      &&
      !((current_sv->init != NULL) &&
	(!sl_slist_empty(current_sv->transitions) || current_sv->all_transitions) &&
	(!sl_slist_empty(current_sv->states))))
    fprintf(stderr,"ERROR: state variable %s badly specified.", current_sv->name);
  else
    sl_add_to_tail(sv_list, current_sv);
}

void add_res_to_res_list(void)
{
  /* Slist *res_list = current_skillset->res_list; */

  /* sl_add_to_tail(res_list, current_res); */
}


Slist *hippo_trace_list;
int hippo_trace_index;

Slist *hippo_user_print_list;
int hippo_user_print_index;

typedef struct string_level {
  int level;
  //  int comp_index;
  char *string;
} StringLevel;

StringLevel *record_trace_string_level(int level, char *string)
{
  StringLevel *sl = MAKE_OBJECT(StringLevel);

  sl->level = level;
  NEWSTR(string, sl->string);
  sl_add_to_tail(hippo_trace_list,sl);

  return sl;
}

StringLevel *record_user_print_string(char *string)
{
  StringLevel *sl = MAKE_OBJECT(StringLevel);
  /* String is actually alredy double quoted...  */

  sl->level = 0;
  NEWSTR(string, sl->string);
  sl_add_to_tail(hippo_user_print_list,sl);

  return sl;
}

void make_trace_string(char **string, const char *format, ...)
{
  va_list va;					
  va_start(va,format);				
  if (vasprintf(string,format,va) == -1)
    warnx("make_trace_string, cannot allocate.");		
  va_end(va);					
}

void output_trace_string(FILE *hf)
{
  fprintf(hf,COPYRIGHT_FILE_HEADER);
  fprintf(hf,
	  "\n\
/*\n\
 * This FIACRE model include file has been automatically generated by bt2f %s.\n\
 * This is the %s.\n\
 * DO NOT EDIT.\n\
 */\n", (bt2f_option[HIPPO]?
	 "online version to be used jointly with Hippo":
	 (bt2f_option[TINA_COMPACT]?
	  "offline abstracted version to be used with the TINA tools":
	  "offline version to be used with the TINA tools")),
	 package_version);

  fprintf(hf,"\n#include \"bt2f-hippo.h\"\n");

  fprintf(hf,"\nStringLevel fiacre_trace_string[%d] = {", hippo_trace_index);

  StringLevel *sl;

  PBoolean first = TRUE;
  sl_loop_through_slist(hippo_trace_list, sl, StringLevel *) {
    fprintf(hf,"%c\n\t{%d, \"%s\"}", (first?' ':','), sl->level, sl->string);
    first = FALSE;
  }

  fprintf(hf,"\n};\n");

  fprintf(hf,"\nStringLevel fiacre_user_print_string[%d] = {", hippo_user_print_index);

  first = TRUE;
  sl_loop_through_slist(hippo_user_print_list, sl, StringLevel *) {
    fprintf(hf,"%c\n\t{%d, \"%s\"}", (first?' ':','), 0, sl->string);
    first = FALSE;
  }

  fprintf(hf,"\n};\n");
}

void output_arg_string(FILE *hargf)
{
  fprintf(hargf,COPYRIGHT_FILE_HEADER);
  fprintf(hargf,
	  "\n\
/*\n\
 * This FIACRE model include file has been automatically generated by bt2f %s.\n \
 * DO NOT EDIT.\n\
 */\n", package_version);

  BTId_Action_Arg * btia;

  sl_loop_through_slist(btidactarg_list, btia,  BTId_Action_Arg *) {
    // if (btnode->type != Action && btnode->type != Condition && !sl_slist_empty(btnode->args)) continue;
    //if (btnode->pass_arg) {	/* probably useless now */
    Sprinter *sp = btia->arg_string;

    SPRINT(sp,150 + strlen(btia->id),
	   sprintf(f,"default:\\\nuc_log_info(\"%s_action undefined arg index.\");\\\n}\\\n}while (0)\n\n\n", btia->id));

    fprintf (hargf,"%s",sprinter_string(sp));
  }

  /* else {			/\* even for btnodes not called, we need to define the macro *\/ */
  /*   static Sprinter *sp = NULL; */

  /*   if (!sp) sp = make_sprinter(0); */

  /*   reset_sprinter(sp); */

  /*   SPRINT(sp,150 + 2*strlen(btnode->id), */
  /* 	   sprintf(f,"\n#define %s_action_arg(index)\\\ndo {\\\nswitch (index) {\\\ndefault:\\\nuc_log_info(\"%s_action undefined arg index.\");\\\n}\\\n}while (0)\n\n\n", */
  /* 		   btia->id, btia->id)); */

  /*   fprintf (hargf,"%s",sprinter_string(sp)); */
  /* } */
}


#define HIPPO_TRACE(level,indent,format,...) do {			\
    if (! bt2f_option[HIPPO]) break;					\
    char *string;							\
    make_trace_string(&string, format, __VA_ARGS__);			\
    record_trace_string_level(level, string);				\
    fprintf(ff,"%s// Trace index: %d, trace level: %d",			\
	    indent, hippo_trace_index, level);				\
    fprintf(ff,"%s// %s",						\
	    indent, string);						\
    fprintf(ff,"%signorep := fiacre_c_print_trace(%d);",		\
	    indent, hippo_trace_index);					\
    hippo_trace_index++;						\
    free(string);							\
  } while (0)

#define USER_PRINT(indent,string) do {					\
    if (! bt2f_option[HIPPO]) break;					\
    char *string2;							\
    int l = strlen(string);						\
    NEWSTR(string+1, string2);						\
    string2[l - 2] ='\0';						\
    record_user_print_string(string2);					\
    fprintf(ff,"%signorep := fiacre_c_print_user_print(%d); // User print index: %d, %s", \
	    indent,hippo_user_print_index,hippo_user_print_index,string); \
    hippo_user_print_index++;						\
    FREE(string2);							\
  } while (0)

#define GRAPHIC_TRACE(indent,btn,status) do {				\
    if (! bt2f_option[HIPPO]) break;					\
    fprintf(ff,"%signorep := fiacre_graphic_trace(%d,%s);", \
	    indent, btn->tcldot, #status);			    \
  } while (0)

unsigned int get_rate_in_ticks_from_args(BTNode *btnode)
{
  Slist *args = btnode->args;
  KeyVal *kv;

  if (! args || sl_slist_empty(args)) {
    fprintf(stderr, "ERROR: expected hz args in arg list of: %s %s_%s.\n",
	    BTNode_Type_Name[btnode->type], btnode->id, btnode->name);
    return 0;
  }
	
  sl_loop_through_slist(args, kv, KeyVal *) {
    Symbol key = kv->key;
    Term *val =  kv->val;

    if (!strcmp(key,"hz")) {
      if (val->type == INTEGER)
	return TICK_PER_RATE(val->u.intval);
      else if (val->type == TT_FLOAT)	
	return TICK_PER_RATE(*(val->u.doubleptr));
    }
  }
  fprintf(stderr, "ERROR: did not find hz args in arg list of: %s %s_%s.\n",
	  BTNode_Type_Name[btnode->type], btnode->id, btnode->name);
  return 0;
}

unsigned int handle_action_arg(BTNode *btnode)
{
  unsigned int arg_index = 0;

  Slist *kvl = btnode->args;

  //  BTNode *btnode1;
  BTId_Action_Arg *btia;

  if (!sl_slist_empty(kvl)) {
    //    Symbol btnode_name = btnode->name;
    Symbol btnode_id = btnode->id;
    PBoolean found = FALSE;

    sl_loop_through_slist(btidactarg_list, btia, BTId_Action_Arg *) {
      if (btia->id == btnode_id) {
	found = TRUE;
	break;
      }
    }

    if (! found) {
      btia = make_btia(btnode_id);
      sl_add_to_head(btidactarg_list, btia);
    }

    Sprinter *sp;
    KeyVal *kv;

    arg_index = action_arg_index++;
    btnode->pass_arg = TRUE;
    if (! (sp = btia->arg_string)) {
	  sp = btia->arg_string = make_sprinter(0);
	  SPRINT(sp,150 + 2*strlen(btnode_id),
		 sprintf(f,"\n#define %s_action_arg(index)\\\ndo {\\\nswitch (index) {\\\ncase 0:\\\nuc_log_info(\"%s_action unexpected empty arg list.\");\\\nbreak;\\\n",  btnode_id, btnode_id));
	}
	//	int pos = 1;

	SPRINT(sp, 30,
	       sprintf(f,"case %d:\\\n", arg_index));

	
	sl_loop_through_slist(kvl, kv, KeyVal *) {
	  Symbol key = kv->key;
	  Term *val =  kv->val;

	  SPRINT(sp,8,
		 sprintf(f,"input."));
	  SPRINT(sp,strlen(key),sprintf(f,"%s",key));

	  SPRINT(sp,3,
		 sprintf(f," = "));
	  sprint_term(sp,val,TRUE);

	  SPRINT(sp,5,
		 sprintf(f,";\\\n"));
	}

	SPRINT(sp, 10,
	       sprintf(f,"break;\\\n"));
	
      }
    /* } */
    /* if (! found) fprintf(stderr,"ERROR: btnode %s not found in the list of btnodes.", btnode_name); */
  /* } */
  return arg_index;
}


/* PBoolean res_list_to_fcr(FILE *ff, Slist *resl) */
/* { */
/*   Res *res; */

/*   fprintf(ff,"\n\n/\* resources types*\/"); */
/*   sl_loop_through_slist(resl, res, Res *) { */
/*     fprintf(ff,"\nconst max_%s: %d..%d is %d", res->name, res->max, res->max, res->max); */
/*     fprintf(ff,"\ntype res_%s is 0..max_%s", res->name, res->name); */
/*   } */
/*   return TRUE; */
/* } */

PBoolean sv_list_to_fcr(FILE *ff, Slist *svl)
{
  Sv *sv;
  Symbol state;

  fprintf(ff,"\n\n/* state variable integer types*/");

  sl_loop_through_slist(svl, sv, Sv *) {

    if (sv->init) continue;

    fprintf(ff,"\ntype sv_%s is %d..%d", sv->name, sv->min, sv->max);
  }

  fprintf(ff,"\n\n/* state variable FSM types*/");

  sl_loop_through_slist(svl, sv, Sv *) {
    PBoolean first;

    if (!sv->init) continue;

    first = TRUE;

    fprintf(ff,"\ntype sv_%s is union", sv->name);
    sl_loop_through_slist(sv->states, state, Symbol ) {
      if (! first) {
	fprintf(ff,"\n\t| %s", state);
      } else {
	fprintf(ff,"\n\t  %s", state);
	first = FALSE;
      }
    }
    fprintf(ff,"\nend\n");
  }

  fprintf(ff,"\n\n/* Process automata controlling the proper transition of resources/SV */");

  sl_loop_through_slist(svl, sv, Sv *) {
    PBoolean first;
    Symbol state2;

    if (!sv->init) continue;

    first = TRUE;

    fprintf(ff,"\n\nprocess sv_%s_automata (&%s:sv_%s) is", sv->name, sv->name, sv->name);

    fprintf(ff,"\n\n\tstates %s%s", sv->init,((! sv->all_transitions)?", error":""));

    sl_loop_through_slist(sv->states, state, Symbol ) {
      if (state != sv->init)
	fprintf(ff,", %s ", state);
    }

    if (bt2f_option[HIPPO]) {
      fprintf(ff,"\n\n\tvar ignorep: nat");
    }

    SymList states2= sl_copy_slist(sv->states);

    SAFE_SL_LOOP_THROUGH_SLIST(sv->states, state, Symbol ) {
      // if ((! sv->all_transitions) && no_transition_from(state,sv->transitions)) continue;

      fprintf(ff,"\n\n\tfrom %s", state);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\tselect");
      first = TRUE;
      SAFE_SL_LOOP_THROUGH_SLIST(states2, state2, Symbol ) { /* this macro is not reentrant... hence the copy */
	if (state2 == state) continue;
	
	if (sv->all_transitions || valid_transition(state, state2, sv->transitions)) {
	  if (! first) {
	    fprintf(ff,"\n\t\t[]");
	  }
	  first = FALSE;
	  fprintf(ff,"\n\t\t\ton (%s = %s);", sv->name, state2);
	  HIPPO_TRACE(4,"\n\t\t\t","State variable '%s', from '%s' to '%s'.", sv->name, state, state2);
	  fprintf(ff,"\n\t\t\tto %s", state2);
	} else {
	  if (! first) {
	    fprintf(ff,"\n\t\t[]");
	  }
	  first = FALSE;
	  fprintf(ff,"\n\t\t\ton (%s = %s);", sv->name, state2);
	  HIPPO_TRACE(0,"\n\t\t\t","State variable '%s', is making a forbiden transiton from '%s' to '%s'.", sv->name, state, state2);
	  fprintf(ff,"\n\t\t\tto error");
	}
      }
      fprintf(ff,"\n\t\tend");
    }
    FREE_SLIST(states2);
  }
  return TRUE;
}

/* PBoolean event_list_to_fcr(FILE *ff, Slist *svl, Slist *evtl) */
/* { */
/*   Sv *sv; */
/*   Event *evt; */
/*   //  Symbol state; */
/*   static Sprinter *sp = NULL; */

/*   fprintf(ff,"\n\n/\* Process controlling the resources transition created by event *\/"); */

/*   sl_loop_through_slist(evtl, evt, Event *) { */
/*     PBoolean first_arg; */
/*     //    Symbol state2; */

/*     //    first = TRUE; */

/*     fprintf(ff,"\n\nprocess event_%s_automata\n\t\t[%s : sync]\n\t\t(", evt->name,  evt->name); */

/*     first_arg = TRUE; */
/*     sl_loop_through_slist(svl, sv, Sv *) { */

/*       if (!first_arg) fprintf(ff,", "); */
/*       else first_arg = FALSE; */
/*       fprintf(ff,"&%s: sv_%s", sv->name, sv->name); */
/*     } */
/*     fprintf(ff,") is"); */

/*     fprintf(ff,"\n\n\tstates start_"); */

/*     if (bt2f_option[HIPPO]) { */
/*       fprintf(ff,"\n\n\tvar ignorep: nat"); */
/*     } */

/*     if (!sp) sp = make_sprinter(0); */
/*     else reset_sprinter(sp); */

/*     fprintf(ff,"\n\n\tfrom start_"); */
/*     fprintf(ff,"\n\t\t%s;", evt->name); */
/*     if (evt->guard) { */
/*       char *reserve = sprint_expr_bt2f_test(sp,evt->guard); */
/*       if (reserve) fprintf(stderr, "ERROR: unexpected resource reserve in test %s\n", sprinter_string(sp)); */
/*       fprintf(ff,"\n\t\ton %s;", sprinter_string(sp)); */
/*       reset_sprinter(sp); */
/*     } */
/*     sprint_list_expr_bt2f_assign(sp,evt->effects); */
/*     fprintf(ff,"\n\t\t%s", sprinter_string(sp)); */
/*     HIPPO_TRACE(4,"\n\t\t","Event automata '%s'.", evt->name); */
/*     fprintf(ff,"\n\tto start_"); */
/*   } */
/*   return TRUE; */
/* } */



PBoolean btnode_list_to_global_declaration(FILE *ff, Slist* btnodel)
{
  BTNode *btnode;
  //  PBoolean first;


  fprintf(ff,"\n\ntype ret_status is union // regular returned status");
  fprintf(ff,"\n\t  success");
  fprintf(ff,"\n\t| failure");
  fprintf(ff,"\n\t| running");
  fprintf(ff,"\n\t| halt_me");
  fprintf(ff,"\n\t| no_ret_status");
  fprintf(ff,"\nend");

  if (bt2f_option[HIPPO]) {
  fprintf(ff,"\n\ntype node_status is union // regular node status");
  fprintf(ff,"\n\t  ns_idle");
  fprintf(ff,"\n\t| ns_ticked");
  fprintf(ff,"\n\t| ns_tick_passed"); /* tick_returned? */
  fprintf(ff,"\n\t| ns_success");
  fprintf(ff,"\n\t| ns_failure");
  fprintf(ff,"\n\t| ns_halt");
  fprintf(ff,"\n\t| ns_halted");
  fprintf(ff,"\n\t| ns_running");
  fprintf(ff,"\nend");
  }

  fprintf(ff,"\n\ntype btnode_caller is union");
  fprintf(ff,"\n\t  None");
  sl_loop_through_slist(btnodel, btnode, BTNode *) {
    if (btnode->type == Action || btnode->type == Condition) continue; /* condition and action cannot be caller */
    fprintf(ff,"\n\t| caller_%s_%s", btnode->id, btnode->name);
  }
  fprintf(ff,"\nend");

  fprintf(ff,"\n\ntype btnode_record is record // record to hold the shared information on a btnode");
  fprintf(ff,"\n\tcaller : btnode_caller, // who called it");
  fprintf(ff,"\n\tArgIndex : nat, // its functional level arguments in");
  //  fprintf(ff,"\n\tnstatus : node_status, // is the invariant being monitored");
  fprintf(ff,"\n\trstatus : ret_status // its last returned status");
  fprintf(ff,"\nend\n");

  int index = 0;
  sl_loop_through_slist(btnodel, btnode, BTNode *) {
    fprintf(ff,"\nconst %s_%s: %d..%d is %d", btnode->id, btnode->name, index, index, index);
    index++;
  }	

  fprintf(ff,"\n\ntype btnode_array is array %d of btnode_record\n", index);

  if (bt2f_option[HIPPO]) {
    fprintf(ff,"\n\n// Global extern functions\n\
extern fiacre_c_print_trace(nat) : nat is\n	\
        print_trace_c\n");
    fprintf(ff,"extern fiacre_c_print_user_print(nat) : nat is\n	\
        print_user_print_c\n");

    fprintf(ff,"extern fiacre_graphic_trace(nat,node_status) : nat is\n	\
        graphic_trace_c\n");
      }

  return TRUE;
}

PBoolean btnode_list_to_fiacre_declaration(FILE *ff, Slist* btnodel)
{
  BTNode *btnode;

  if (bt2f_option[HIPPO]) {
    Slist *declared = sl_make_slist();
    fprintf(ff,"\n\n// Hippo tasks for all basic btnodes.");
    sl_loop_through_slist(btnodel, btnode, BTNode *) {
      if (btnode->type == SetSV) {
	if (!sl_in_slist(declared, btnode->id)) {
	  fprintf(ff,"\nextern Fiacre_%s_%s_extern () : sv_%s is", BTNode_Type_Name[btnode->type], btnode->id, btnode->sv);
	  fprintf(ff,"\n\t\tFiacre_%s_%s_c", BTNode_Type_Name[btnode->type], btnode->id);
	  sl_add_to_tail(declared, btnode->id);
	}
      }

      if (!sl_in_slist(declared, btnode->id)) {
	if (btnode->type == Condition) {
	    fprintf(ff,"\nextern  Fiacre_%s_%s_ext (btnode_record) : ret_status is", BTNode_Type_Name[btnode->type], btnode->id);
	    fprintf(ff,"\n\t\tFiacre_%s_%s_c", BTNode_Type_Name[btnode->type], btnode->id);
	    sl_add_to_tail(declared, btnode->id);
	}
	if (btnode->type == Action) { /* conditions are now allowed to be "running", should they be task? */
	    fprintf(ff,"\ntask Fiacre_%s_%s_task (btnode_record) : ret_status is", BTNode_Type_Name[btnode->type], btnode->id);
	    fprintf(ff,"\n\t\tFiacre_%s_%s_c", BTNode_Type_Name[btnode->type], btnode->id);
	    fprintf(ff,"\nextern Fiacre_%s_%s_halt (btnode_record) : bool is", BTNode_Type_Name[btnode->type], btnode->id);
	    fprintf(ff,"\n\t\tFiacre_%s_%s_halt_c", BTNode_Type_Name[btnode->type], btnode->id);
	    sl_add_to_tail(declared, btnode->id);
	}
      }
    }
    FREE_SLIST(declared);
  }

  return TRUE;
}


PBoolean btnode_list_to_fcr(FILE *ff, Slist *svl, //Slist *resl,
			    Slist* btnodel)
{
  Sv *sv;
  BTNode *btnode;

  static Sprinter *sp = NULL;
  if (!sp) sp = make_sprinter(0);

  int index = 0;

  sl_loop_through_slist(btnodel, btnode, BTNode *) {

    const BTNode *btnode1, *btnode2, *btnodefirst;
    static Sprinter *sp_args = NULL;
    char *arg_string;

    if (!sl_slist_empty(btnode->args)) {

      Slist *kvl = btnode->args;
      KeyVal *kv;
      PBoolean first;

      if (! sp_args)
	sp_args = make_sprinter(0);
      else
	reset_sprinter(sp_args);

      SPRINT(sp_args,2,sprintf(f," ("));

      first = TRUE;
      sl_loop_through_slist(kvl, kv, KeyVal *) {
	Symbol key = kv->key;
	Term *val =  kv->val;
	
	
	SPRINT(sp_args,strlen(key)+3,sprintf(f,"%s:%s ",(!first?" ":""),key));
	first = FALSE;

	sprint_term(sp_args,val,TRUE);

	SPRINT(sp_args,1, sprintf(f," "));
      }

      SPRINT(sp_args,2,sprintf(f,") "));

      arg_string = sprinter_string(sp_args);
    } else {
      arg_string = " ";
    }

    fprintf(ff,"\n\nprocess btnode_%s_%s (",
	    (btnode->id?btnode->id:"Undefined"),
	    (btnode->name?btnode->name:"Undefined"));

    fprintf(ff,"&btnode: btnode_array");
    if (btnode->type == Action || btnode->type == Condition || btnode->type == Eval || btnode->type == SetSV) {
      sl_loop_through_slist(svl, sv, Sv *) {
	fprintf(ff,", &%s: sv_%s", sv->name, sv->name);
      }
    }
    fprintf(ff,") is");

    /* States */
    fprintf(ff,"\n\n\tstates");
    fprintf(ff," start_, tick_node,");

    if (btnode->type != AlwaysFailure && btnode->type != ForceFailure && btnode->type != KeepRunningUntilFailure) {
      fprintf(ff," success,");
    }
    
    if (btnode->type != SetSV && btnode->type != AlwaysSuccess && btnode->type != ForceSuccess &&
	! (btnode->type == Eval && (ASSIGN_P(btnode->expr)))) {
      fprintf(ff," failure,");
    }
    
    if (btnode->type != Condition) {
      fprintf(ff," halt, halted,");
    }

    if (! (btnode->type ==  BehaviorTree || btnode->type == Condition || btnode->type == Action ||
	   btnode->type == Eval || btnode->type == SetSV)) {
      fprintf(ff," halt_wait,");
    }

    if (btnode->type == Parallel || btnode->type == ParallelAll) {
      fprintf(ff,"\n\t\trunning, %s_%s_par, %s_%s_par_done,", btnode->id, btnode->name, btnode->id, btnode->name);
      if (btnode->wait) {
	fprintf(ff," failure_wait, failure_wait2,");
	if (btnode->type == Parallel) /* ParallelAll do not have success_wait */
	  fprintf(ff," success_wait, success_wait2,");
      }
    } else {
      if (btnode->type == Action) {
	if (btnode->PossibleRet == SFR || btnode->PossibleRet == SR ||
	    btnode->PossibleRet == FR || btnode->PossibleRet == R) {
	  fprintf(ff," running,");
	}
      } else {
	if (btnode->type != Eval && btnode->type != Condition && btnode->type != SetSV) {
	  fprintf(ff," running,");
	}
      }

      if ((bt2f_option[HIPPO] && btnode->type != SetSV && btnode->type != Eval) ||
	  (!bt2f_option[HIPPO] && btnode->type != Condition && btnode->type != Action &&
	   btnode->type != SetSV && btnode->type != Eval))
	fprintf(ff," error,");

      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\t%s_%s, %s_%s_done,", btnode1->id, btnode1->name, btnode1->id, btnode1->name);
      }
    }

    if (bt2f_option[HIPPO]) {
      if (btnode->type == Condition || btnode->type == Action) {
	fprintf(ff,"\n\t\t%s_%s, dispatch,",  BTNode_Type_Name[btnode->type], btnode->id);
      }
      if (btnode->type == Action) {
	fprintf(ff," %s_%s_sync,",  BTNode_Type_Name[btnode->type], btnode->id);
      }
    }
  
    fprintf(ff,"\n\t\tdone, ether");

    /* Local vars */
    fprintf(ff,"\n\n\tvar ignorep: nat, ignoreb: bool");

    if (btnode->type == Repeat) {
      fprintf(ff,",\n\t\trep: 0..%d := 0, repeat: %d..%d := %d", btnode->u.repeat,
	      btnode->u.repeat, btnode->u.repeat, btnode->u.repeat);
    }

    if (btnode->type == Parallel) {
      fprintf(ff,",\n\t\tsuccess_so_far: 0..%d := 0, failed_so_far: 0..%d := 0",
	      sl_slist_length(btnode->children), sl_slist_length(btnode->children));
    }

    if (btnode->type == PipelineSequence || btnode->type == Parallel || btnode->type == RoundRobin ||
	btnode->type == ParallelAll || btnode->type == ReactiveSequence) {
      fprintf(ff,",\n\t\trunningb:bool := false");
      fprintf(ff,",\n\t\thalt_called_by_me:bool := false");
      fprintf(ff,",\n\t\tforce_success:bool := false");
    }

    if (btnode->type != Condition && bt2f_option[HIPPO]) {
      fprintf(ff,",\n\t\treport_halted:bool := false");
    }

    if (btnode->type == RetryUntilSuccessful) {
      fprintf(ff,",\n\t\tretry: 0..%d := 0, attempts: nat := %d", btnode->u.num_attempts, btnode->u.num_attempts);
    }

    if (btnode->type == Sequence || btnode->type == ReactiveSequence ||
	btnode->type == RoundRobin || btnode->type == SequenceWithMemory) {
      fprintf(ff,",\n\t\tnext_seq: 1..%d := 1", sl_slist_length(btnode->children)+1);
    }

    if (btnode->type == Fallback || btnode->type == ReactiveFallback) {
      fprintf(ff,",\n\t\tnext_fb: 1..%d := 1", sl_slist_length(btnode->children));
    }

    if (btnode->type == Recovery) {
      fprintf(ff,",\n\t\tnext_rec: 1..%d := 1", sl_slist_length(btnode->children));
    }

    if (btnode->type == PipelineSequence) {
      fprintf(ff,",\n\t\tlast_succ: 1..%d := 1", sl_slist_length(btnode->children));
    }

    if (btnode->type ==  RateController) {
      fprintf(ff,",\n\t\trate: 1..%d := 1", get_rate_in_ticks_from_args(btnode));
    }

    if (btnode->type == RoundRobin) {
      fprintf(ff,",\n\t\tfailed: 0..%d := 0", sl_slist_length(btnode->children));
    }

    if ((btnode->type == Condition || btnode->type == Action) && bt2f_option[HIPPO]) {
      fprintf(ff,", ret_val: ret_status");
    }

    /* Transitions */
    fprintf(ff,"\n\n\tfrom start_");
    fprintf(ff,"\n\t\twait [0,0];");
    fprintf(ff,"\n\t\ton (btnode[%s_%s].caller <> None);", btnode->id, btnode->name);

    /* Halt or not? */
    if (btnode->type != Condition) {
      if (bt2f_option[HIPPO])
	fprintf(ff,"\n\t\treport_halted := false;");
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = halt_me) then", btnode->id, btnode->name);
      if (bt2f_option[HIPPO])
	fprintf(ff,"\n\t\t\treport_halted := true;");
      fprintf(ff,"\n\t\t\tto halt");
      fprintf(ff,"\n\t\tend;");
    }
    HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' has been called%s",
		BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
    fprintf(ff,"\n\t\tbtnode[%s_%s].rstatus := no_ret_status;", btnode->id, btnode->name);
    fprintf(ff,"\n\tto tick_node");
    
    if (btnode->type != Condition) {
      fprintf(ff,"\n\n\tfrom halt");
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' is being halted%s",
		  BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_halt);
    }

    /* How do we halt... */
    switch(btnode->type) {
    case Action:
      if (bt2f_option[HIPPO]) {
	fprintf(ff,"\n\t\tignoreb := Fiacre_%s_%s_halt (btnode[%s_%s]);",
		BTNode_Type_Name[btnode->type], btnode->id, btnode->id, btnode->name);
      }
      fprintf(ff,"\n\tto halted");
      break;

    case Sequence:
    case ReactiveSequence:
    case SequenceWithMemory:
    case Recovery:
    case Fallback:
    case RoundRobin:
    case ReactiveFallback:
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' halting %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\t\tbtnode[%s_%s].rstatus := halt_me;", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
	fprintf(ff,"\n\t\t\tto halt_wait");
	fprintf(ff,"\n\t\tend;");
      }
      fprintf(ff,"\n\t\tto halted");

      fprintf(ff,"\n\n\tfrom halt_wait");
      fprintf(ff,"\n\t\ton (");
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\t\t(btnode[%s_%s].caller = None) and", btnode1->id, btnode1->name);
      }
      fprintf(ff,"\n\t\t\ttrue);");
      fprintf(ff,"\n\t\tto halted");
      break;

      /* Possibly multiple running children...  */
    case Parallel:
    case ParallelAll:
    case PipelineSequence:	
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' halting %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\t\tbtnode[%s_%s].rstatus := halt_me;", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s", btnode1->id, btnode1->name, btnode->id, btnode->name);
	fprintf(ff,"\n\t\tend;");
      }
      fprintf(ff,"\n\t\tto halt_wait");

      fprintf(ff,"\n\n\tfrom halt_wait");
      fprintf(ff,"\n\t\ton (");	/* wait multiple */
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\t\t(btnode[%s_%s].caller = None) and", btnode1->id, btnode1->name);
      }
      fprintf(ff,"\n\t\t\ttrue);");
      fprintf(ff,"\n\tto halted");

      break;

      /* Only one child possibly running...  */
    case ForceFailure:
    case ForceSuccess:
    case Inverter:
    case KeepRunningUntilFailure:
    case RetryUntilSuccessful:
    case RateController:
    case Repeat:
      if (btnode->type ==  RateController) {
	fprintf(ff,"\n\t\trate := 1;");
      }
      btnode1 = sl_get_slist_head(btnode->children);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' halting %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\t\tbtnode[%s_%s].rstatus := halt_me;", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      fprintf(ff,"\n\t\t\tto halt_wait");
      fprintf(ff,"\n\t\tend;");
      fprintf(ff,"\n\t\tto halted");

      fprintf(ff,"\n\n\tfrom halt_wait");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      fprintf(ff,"\n\tto halted");
      break;

    case BehaviorTree:
    case Eval:
    case SetSV:
      fprintf(ff,"\n\tto halted");
      break;			/* Nothing, these cannot be running */

    case Condition:
      break;

    default:
      fprintf(stderr, "ERROR: unexpected BT node type to halt: %s\n", BTNode_Type_Name[btnode->type]);
      fprintf(ff,"\n\tto halted");
      break;
    }
    /* Done halting */

    /* The beef. */
    fprintf(ff,"\n\n\tfrom tick_node");
    fprintf(ff,"\n\t\twait [%d,%d];",(htick == NODE?1:
				      (htick == ACTION && (btnode->type == Action ||
							   btnode->type == Condition ||
							   btnode->type == Eval ||
							   btnode->type == SetSV ||
							   sl_in_slist(bt_list,btnode))?1:
				       (htick == BT && sl_in_slist(bt_list,btnode)?1:0))),
	    (htick == NODE?1:
	     (htick == ACTION && (btnode->type == Action ||
				  btnode->type == Condition ||
				  btnode->type == Eval ||
				  btnode->type == SetSV ||
				  sl_in_slist(bt_list,btnode))?1:
	      (htick == BT && sl_in_slist(bt_list,btnode)?1:0))));
    GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);

    PBoolean first = TRUE;


    switch(btnode->type) {

    case AlwaysSuccess:
      fprintf(ff,"\n\t\tto success");
      break;

    case AlwaysFailure:
      fprintf(ff,"\n\t\tto failure");
      break;

    case Sequence:
    case ReactiveSequence:
    case SequenceWithMemory:
      index = 1;
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (next_seq = %d) then to %s_%s end;", index, btnode1->id, btnode1->name);
	index++;
      }

      index = 1;
      btnode1 = sl_get_slist_next(btnode->children, NULL);
      btnode2 = sl_get_slist_next(btnode->children, btnode1);

      fprintf(ff,"\n\t\tto error");
      //      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);

      while (btnode1) {
	
	fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
	fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
	if (btnode2)
	  fprintf(ff,"\n\t\t\tto %s_%s", btnode2->id, btnode2->name);
	else {
	  fprintf(ff,"\n\t\t\tnext_seq := 1;"); /* reset it */
	  fprintf(ff,"\n\t\t\tto success"); /* The last one, return success */
	}
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
	if (btnode->type == SequenceWithMemory) {
	  fprintf(ff,"\n\t\t\tnext_seq := %d;", index);
	  fprintf(ff,"\n\t\t\tto failure");
	} else {
	  fprintf(ff,"\n\t\t\tnext_seq := 1;");
	  if (btnode->type == ReactiveSequence && btnode->halt) { /* Halt the running right branch if any... */
	    fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	    fprintf(ff,"\n\t\t\tto halt");
	  } else {
	    fprintf(ff,"\n\t\t\tto failure");
	  }
	}
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	if (btnode->type == Sequence || btnode->type == SequenceWithMemory) {
	  fprintf(ff,"\n\t\t\tnext_seq := %d;", index);
	} else {		/* ReactiveSequence */
	  fprintf(ff,"\n\t\t\tnext_seq := 1;");
	}
	fprintf(ff,"\n\t\t\tto running");
	fprintf(ff,"\n\t\telse");
	fprintf(ff,"\n\t\t\tto error");
	fprintf(ff,"\n\t\tend");

	btnode1 = btnode2;
	btnode2 = sl_get_slist_next(btnode->children, btnode1);
	index++;
      }

      break;

    case PipelineSequence:

      index = 1;
      btnode1 = sl_get_slist_next(btnode->children, NULL);
      btnode2 = sl_get_slist_next(btnode->children, btnode1);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);

      while (btnode1) {
	
	fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
	fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
	if (btnode2) {
	  fprintf(ff,"\n\t\t\tif (%d = last_succ) then last_succ := last_succ + 1 end;", index);
	  fprintf(ff,"\n\t\t\tto %s_%s", btnode2->id, btnode2->name);
	} else {
	  fprintf(ff,"\n\t\t\tlast_succ := 1;"); /* Last one, reset */
	  fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	  fprintf(ff,"\n\t\t\tforce_success := true;");
	  fprintf(ff,"\n\t\t\tto halt"); /* The last one, return success but we halt any running first */
	}
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tlast_succ := 1;"); /* failure, reset */
	fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	fprintf(ff,"\n\t\t\tto halt");
	
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	if (btnode2) {
	  fprintf(ff,"\n\t\t\tif (last_succ >= %d) then to %s_%s end;", /* We already went further down... keep going. */
		  index, btnode2->id, btnode2->name);
	}
	fprintf(ff,"\n\t\t\tto running");
	fprintf(ff,"\n\t\telse");
	fprintf(ff,"\n\t\t\tto error");
	fprintf(ff,"\n\t\tend");

	btnode1 = btnode2;
	btnode2 = sl_get_slist_next(btnode->children, btnode1);
	index++;
      }

      break;

    case Fallback:
    case ReactiveFallback:
      index = 1;
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (next_fb = %d) then to %s_%s end;", index, btnode1->id, btnode1->name);
	index++;
      }

      fprintf(ff,"\n\t\tto error");
      //      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);

      index = 1;
      btnode1 = sl_get_slist_next(btnode->children, NULL);
      btnode2 = sl_get_slist_next(btnode->children, btnode1);

      while (btnode1) {
	
	fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
	fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tnext_fb := 1;"); /* reset it */
	fprintf(ff,"\n\t\t\tto success");
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
	if (btnode2)
	  fprintf(ff,"\n\t\t\tto %s_%s", btnode2->id, btnode2->name);
	else {					/* last one */
	  fprintf(ff,"\n\t\t\tnext_fb := 1;"); /* reset it */
	  fprintf(ff,"\n\t\t\tto failure");
	}
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	if (btnode->type == Fallback) {
	  fprintf(ff,"\n\t\t\tnext_fb := %d;", index);
	} else {		/* ReactiveFallback */
	  fprintf(ff,"\n\t\t\tnext_fb := 1;");
	}
	fprintf(ff,"\n\t\t\tto running");
	fprintf(ff,"\n\t\telse");
	fprintf(ff,"\n\t\t\tto error");
	fprintf(ff,"\n\t\tend");

	btnode1 = btnode2;
	btnode2 = sl_get_slist_next(btnode->children, btnode1);
	index++;
      }
      break;

    case RoundRobin:
      index = 1;
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (next_seq = %d) then to %s_%s end;", index, btnode1->id, btnode1->name);
	index++;
      }

      index = 1;
      btnodefirst = btnode1 = sl_get_slist_next(btnode->children, NULL);
      btnode2 = sl_get_slist_next(btnode->children, btnode1);

      fprintf(ff,"\n\t\tto error");
      //      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);

      while (btnode1) {
	
	fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
	fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\twait [0,0];");
	fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
	GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
	if (btnode2)
	  fprintf(ff,"\n\t\t\tnext_seq := %d;", index + 1); /* next node, next time */
	else
	  fprintf(ff,"\n\t\t\tnext_seq := 1;"); /*restart from beginning */
	fprintf(ff,"\n\t\t\tfailed := 0;"); /* reset failure counter */
	fprintf(ff,"\n\t\t\tforce_success := true;");
	fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	fprintf(ff,"\n\t\t\tto halt");/* The last one, return success after halting running*/

	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tfailed := failed + 1;"); /* reset failure counter */
	if (btnode2)
	  fprintf(ff,"\n\t\t\tnext_seq := %d;", index + 1); /* next node, next time */
	else
	  fprintf(ff,"\n\t\t\tnext_seq := 1;"); /*restart from beginning */
	fprintf(ff,"\n\t\t\tif (failed = %d) then", sl_slist_length(btnode->children)); /* reset failure counter */
	fprintf(ff,"\n\t\t\t\tfailed := 0;"); /* reset failure counter */
	fprintf(ff,"\n\t\t\t\tto failure");
	fprintf(ff,"\n\t\t\telse"); /* not all failed yet */
	if (btnode2)
	  fprintf(ff,"\n\t\t\t\tto %s_%s", btnode2->id, btnode2->name);
	else
	  fprintf(ff,"\n\t\t\t\tto %s_%s", btnodefirst->id, btnodefirst->name);
	fprintf(ff,"\n\t\t\tend"); /* reset failure counter */
	
	fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	fprintf(ff,"\n\t\t\tto running");
	fprintf(ff,"\n\t\telse");
	fprintf(ff,"\n\t\t\tto error");
	fprintf(ff,"\n\t\tend");

	btnode1 = btnode2;
	btnode2 = sl_get_slist_next(btnode->children, btnode1);
	index++;
      }

      break;
	
    case KeepRunningUntilFailure:
      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case Inverter:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;
	
    case RateController:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\tif (rate = 1) then");
      fprintf(ff,"\n\t\t\trate := %d;", get_rate_in_ticks_from_args(btnode));
      HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\t\tto %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\trate := rate - 1;");
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\tend");

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case ForceSuccess:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case ForceFailure:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case Repeat:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\trep := rep + 1 ;");
      fprintf(ff,"\n\t\t\tif (rep = repeat) then");
      fprintf(ff,"\n\t\t\t\trep := 0;");
      fprintf(ff,"\n\t\t\t\tto success");
      fprintf(ff,"\n\t\t\tend;");
      fprintf(ff,"\n\t\t\tto tick_node");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\trep := 0;");
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case BehaviorTree:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case Eval:
      reset_sprinter(sp);
      sprint_expr(sp, btnode->expr, TRUE);
      HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' evaluating expression %s",
		  BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, sprinter_string(sp));

      if (ASSIGN_P(btnode->expr)) {
	TermList ts = btnode->expr->terms;
	Term *term1 = NULL;
	Term *term2 = NULL;

	if (ts && (sl_slist_length(ts) == 2)){
	  term1 = (Term *) sl_get_slist_pos(ts, 1);
	  term2 = (Term *) sl_get_slist_pos(ts, 2);
	}
	
	if (term1->type == TT_ATOM) { /* expect an atom (the variable) */
	  reset_sprinter(sp);
	  sprint_term(sp, term1, FALSE);
	  SPRINT(sp,5,sprintf(f," := "));
	  sprint_term(sp, term2, FALSE);
	  fprintf(ff,"\n\t\t// will assign: %s", sprinter_string(sp));
	  HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' assign '%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, sprinter_string(sp));
	  fprintf(ff,"\n\t\t%s;", sprinter_string(sp));
	  fprintf(ff,"\n\t\tto success");
	} else {
	  fprintf(stderr, "ERROR: expected a variable name before an assignment.");
	}
      } else {
	reset_sprinter(sp);
	char *reserve = sprint_expr_bt2f_test(sp, btnode->expr);
	if (reserve) fprintf(stderr, "ERROR: unexpected resource reserve in Eval BT: %s\n", sprinter_string(sp));
	fprintf(ff,"\n\t\tif %s then",  sprinter_string(sp));
	fprintf(ff,"\n\t\t\tto success");
	fprintf(ff,"\n\t\telse");
	fprintf(ff,"\n\t\t\tto failure");
	fprintf(ff,"\n\t\tend");
      }
      break;

    case SetSV:
      if (bt2f_option[HIPPO]) {
	HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' Setting State Variable %s.", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, btnode->sv);
	fprintf(ff,"\n\t\t%s := Fiacre_%s_%s_extern();", btnode->sv,  BTNode_Type_Name[btnode->type], btnode->id);
	fprintf(ff,"\n\t\tto success");
      } else {
	Sv *sv = NULL;
	sl_loop_through_slist(sv_list, sv, Sv *) {
	  if (!strcmp(sv->name,btnode->sv)) break;
	}
	if (sv && sv->init) {	// automata
	  Symbol state;
	  fprintf(ff,"\n\t\tselect");
	  PBoolean first = TRUE;
	  sl_loop_through_slist(sv->states, state, Symbol ) {
	    if (! first) {
	      fprintf(ff,"\n\t\t[]");
	    }
	    first = FALSE;
	    fprintf(ff,"\n\t\t\t%s := %s", btnode->sv, state);
	  }
	  fprintf(ff,"\n\t\tend;");

	} else {
	  fprintf(ff,",\n\t\t%s := any", btnode->sv);
	}
	
	fprintf(ff,"\n\t\tto success");
      }
      break;

    case Action:
    case Condition:
      if (bt2f_option[HIPPO]) {
	if (btnode->type == Action) {

	  fprintf(ff,"\n\t\tto %s_%s",  BTNode_Type_Name[btnode->type], btnode->id);

	  fprintf(ff,"\n\n\tfrom %s_%s",  BTNode_Type_Name[btnode->type], btnode->id);
	  if (bt2f_option[HIPPO] && !sl_slist_empty(btnode->args)) {
	    int arg_index = handle_action_arg(btnode);
	    fprintf(ff,"\n\t\t// synthesized action arg index  %d", arg_index);
            fprintf(ff,"\n\t\tbtnode[%s_%s].ArgIndex := %d;",  btnode->id,  btnode->name , arg_index);
	  }

	
	  HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' calling its action (start task)", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);
	  fprintf(ff,"\n\t\tstart Fiacre_%s_%s_task (btnode[%s_%s]);",  BTNode_Type_Name[btnode->type], btnode->id,  btnode->id, btnode->name);
	  fprintf(ff,"\n\tto %s_%s_sync",  BTNode_Type_Name[btnode->type], btnode->id);

	  fprintf(ff,"\n\n\tfrom %s_%s_sync",  BTNode_Type_Name[btnode->type], btnode->id);
	  fprintf(ff,"\n\t\tsync Fiacre_%s_%s_task ret_val;", BTNode_Type_Name[btnode->type], btnode->id);
	  HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' returned (sync task)", BTNode_Type_Name[btnode->type], btnode->id, btnode->name);
	  fprintf(ff,"\n\t\tto dispatch");
	
	  fprintf(ff,"\n\n\tfrom dispatch");
	  fprintf(ff,"\n\t\twait [0,0];");
	  fprintf(ff,"\n\t\tif (ret_val = success) then");
	  fprintf(ff,"\n\t\t\tto success");
	  fprintf(ff,"\n\t\telsif (ret_val = failure) then");
	  fprintf(ff,"\n\t\t\tto failure");
	  fprintf(ff,"\n\t\telsif (ret_val = running) then");
	  fprintf(ff,"\n\t\t\tto running");
	  fprintf(ff,"\n\t\telse");
	  fprintf(ff,"\n\t\t\tto error // a priori unreachable");
	  fprintf(ff,"\n\t\tend");
	} else {			/* Condition... simpler */

	  fprintf(ff,"\n\t\tto %s_%s",  BTNode_Type_Name[btnode->type], btnode->id);

	  fprintf(ff,"\n\n\tfrom %s_%s",  BTNode_Type_Name[btnode->type], btnode->id);
	  if (bt2f_option[HIPPO] && !sl_slist_empty(btnode->args)) {
	    int arg_index = handle_action_arg(btnode);
	    fprintf(ff,"\n\t\t// synthesized action arg index  %d", arg_index);
            fprintf(ff,"\n\t\tbtnode[%s_%s].ArgIndex := %d;",  btnode->id,  btnode->name , arg_index);
	  }

	
	  HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' calling its condition external", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);
	  fprintf(ff,"\n\t\tret_val := Fiacre_%s_%s_ext (btnode[%s_%s]);",  BTNode_Type_Name[btnode->type], btnode->id,  btnode->id, btnode->name);
	  fprintf(ff,"\n\t\tto dispatch");
	
	  fprintf(ff,"\n\n\tfrom dispatch");
	  fprintf(ff,"\n\t\twait [0,0];");
	  fprintf(ff,"\n\t\tif (ret_val = success) then");
	  fprintf(ff,"\n\t\t\tto success");
	  fprintf(ff,"\n\t\telsif (ret_val = failure) then");
	  fprintf(ff,"\n\t\t\tto failure");
	  fprintf(ff,"\n\t\telse");
	  fprintf(ff,"\n\t\t\tto error // a priori unreachable");
	  fprintf(ff,"\n\t\tend");
	}
      } else {

	fprintf(ff,"\n\t\tselect");
	if (btnode->PossibleRet == SFR || btnode->PossibleRet == SF || btnode->PossibleRet == SR || btnode->PossibleRet == S) {
	  if (! first)
	    fprintf(ff,"\n\t\t[]");
	  fprintf(ff,"\n\t\t\tto success");
	  first = FALSE;
	}
	if (btnode->PossibleRet == SFR || btnode->PossibleRet == SF || btnode->PossibleRet == FR || btnode->PossibleRet == F) {
	  if (! first)
	    fprintf(ff,"\n\t\t[]");
	  fprintf(ff,"\n\t\t\tto failure");
	  first = FALSE;
	}
	if (btnode->type != Condition && (btnode->PossibleRet == SFR || btnode->PossibleRet == SR || btnode->PossibleRet == FR || btnode->PossibleRet == R)) { /* Condition should not return running */
	  if (! first)
	    fprintf(ff,"\n\t\t[]");
	  fprintf(ff,"\n\t\t\tto running");
	  first = FALSE;
	}
	fprintf(ff,"\n\t\tend");
      }
      break;

    case Parallel:
    case ParallelAll:

      fprintf(ff,"\n\tto %s_%s_par", btnode->id, btnode->name);

      fprintf(ff,"\n\n\tfrom %s_%s_par", btnode->id, btnode->name);
      fprintf(ff,"\n\t\twait [0,0];");

      fprintf(ff,"\n\t\tif (runningb) then");

      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	HIPPO_TRACE(4,"\n\t\t\t\t","Btnode %s '%s_%s' calling again %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\t\t\tbtnode[%s_%s].caller := caller_%s_%s", btnode1->id, btnode1->name, btnode->id, btnode->name);
	fprintf(ff,"\n\t\t\tend;");
      }
      fprintf(ff,"\n\t\t\tnull");
      fprintf(ff,"\n\t\telse");
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      }
      fprintf(ff,"\n\t\t\tnull");
      fprintf(ff,"\n\t\tend;");

      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);

      fprintf(ff,"\n\t\tto %s_%s_par_done", btnode->id, btnode->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s_par_done", btnode->id, btnode->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (");
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\t\t(btnode[%s_%s].caller = None) and", btnode1->id, btnode1->name);
      }
      fprintf(ff,"\n\t\t\ttrue);");
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);

      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      }

      if (btnode->type == ParallelAll) {
	  fprintf(ff,"\n\t\tif (");
	  sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	    fprintf(ff,"\n\t\t\t(btnode[%s_%s].rstatus = success) and", btnode1->id, btnode1->name);
	  }
	  fprintf(ff,"\n\t\t\ttrue) then");
	  fprintf(ff,"\n\t\t\t\trunningb := false;");
	  fprintf(ff,"\n\t\t\tto success");
	  fprintf(ff,"\n\t\telsif (");
	  sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	    fprintf(ff,"\n\t\t\t(btnode[%s_%s].rstatus = failure) or", btnode1->id, btnode1->name);
	  }
	  fprintf(ff,"\n\t\t\tfalse) then");
	  fprintf(ff,"\n\t\t\t\trunningb := false;");
	  if (btnode->wait) /* if one failure then failure */
	    fprintf(ff,"\n\t\t\tto failure_wait"); /* wait for running left */
	  else if (btnode->halt) {
	    fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	    fprintf(ff,"\n\t\t\tto halt"); /* halt the running left */
	  } else
	    fprintf(ff,"\n\t\t\tto failure");
	  fprintf(ff,"\n\t\telse "); /* at least still one running and none failed */
      } else {

	fprintf(ff,"\n\t\tsuccess_so_far := 0; failed_so_far := 0;");

	sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	  fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then success_so_far := success_so_far + 1", btnode1->id, btnode1->name);
	  fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then failed_so_far := failed_so_far + 1 end; ", btnode1->id, btnode1->name);
	}

	fprintf(ff,"\n\t\tif (success_so_far >= %d) then", btnode->u.needed_success);
	HIPPO_TRACE(3,"\n\t\t\t","Btnode %s '%s_%s' has enough success: >= %d", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, btnode->u.needed_success);
	if (btnode->wait)
	  fprintf(ff,"\n\t\t\tto success_wait"); /* wait for running left */
	else if (btnode->halt) { /* if one failure then failure */
	  fprintf(ff,"\n\t\t\thalt_called_by_me := true;");
	  fprintf(ff,"\n\t\t\tforce_success := true;");
	  fprintf(ff,"\n\t\t\tto halt"); /* halt the running left */
	} else
	  fprintf(ff,"\n\t\t\tto success");
	fprintf(ff,"\n\t\telsif (failed_so_far > %d) then", sl_slist_length(btnode->children) -  btnode->u.needed_success);
	HIPPO_TRACE(3,"\n\t\t\t","Btnode %s '%s_%s' has too many failure(s): > %d", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, sl_slist_length(btnode->children) - btnode->u.needed_success);
	if (btnode->wait) /* if one failure then failure */
	  fprintf(ff,"\n\t\t\tto failure_wait"); /* wait for running left */
	else
	  fprintf(ff,"\n\t\t\tto failure");
	//      fprintf(ff,"\n\t\t\tto failure");
	fprintf(ff,"\n\t\telse");
      }
      fprintf(ff,"\n\t\t\trunningb := true;");
      fprintf(ff,"\n\t\t\tto running");

      fprintf(ff,"\n\t\tend");

      if (btnode->wait) {
	if (btnode->type == Parallel) { /* ParallelAll do not have success_wait */
	  fprintf(ff,"\n\n\tfrom success_wait"); /* wait for running left */
	  fprintf(ff,"\n\t\twait [0,0];");
	  sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	    fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	    HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' success but calling again to wait running %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	    fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s", btnode1->id, btnode1->name, btnode->id, btnode->name);
	    fprintf(ff,"\n\t\t\tend;");
	  }
	  GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	  fprintf(ff,"\n\tto success_wait2"); /* wait for running left */
	  fprintf(ff,"\n\n\tfrom success_wait2"); /* wait for running left */
	  fprintf(ff,"\n\t\twait [0,0];");
	  fprintf(ff,"\n\t\ton (");
	  sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	    fprintf(ff,"\n\t\t\t(btnode[%s_%s].caller = None) and", btnode1->id, btnode1->name);
	  }
	  fprintf(ff,"\n\t\t\ttrue);");
	  GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	  fprintf(ff,"\n\t\tif (");
	  sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	    fprintf(ff,"\n\t\t\t(btnode[%s_%s].rstatus = running) or", btnode1->id, btnode1->name);
	  }
	  fprintf(ff,"\n\t\t\tfalse) then");
	  fprintf(ff,"\n\t\t\tto success_wait");
	  fprintf(ff,"\n\t\telse");
	  fprintf(ff,"\n\t\t\tto success");
	  fprintf(ff,"\n\t\tend");
	}
	fprintf(ff,"\n\n\tfrom failure_wait"); /* wait for running left */
	fprintf(ff,"\n\t\twait [0,0];");
	sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	  fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
	  HIPPO_TRACE(4,"\n\t\t\t","Btnode %s '%s_%s' failed but calling again to wait running %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
	  fprintf(ff,"\n\t\t\tbtnode[%s_%s].caller := caller_%s_%s", btnode1->id, btnode1->name, btnode->id, btnode->name);
	  fprintf(ff,"\n\t\tend;");
	}
	GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
	fprintf(ff,"\n\tto failure_wait2"); /* wait for running left */
	fprintf(ff,"\n\n\tfrom failure_wait2"); /* wait for running left */
	fprintf(ff,"\n\t\twait [0,0];");
	fprintf(ff,"\n\t\ton (");
	sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	  fprintf(ff,"\n\t\t\t(btnode[%s_%s].caller = None) and", btnode1->id, btnode1->name);
	}
	fprintf(ff,"\n\t\t\ttrue);");
	GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
	fprintf(ff,"\n\t\tif (");
	sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	  fprintf(ff,"\n\t\t\t(btnode[%s_%s].rstatus = running) or", btnode1->id, btnode1->name);
	  }
	  fprintf(ff,"\n\t\t\tfalse) then");
	  fprintf(ff,"\n\t\t\t\tto failure_wait");
	  fprintf(ff,"\n\t\t\telse");
	  fprintf(ff,"\n\t\t\t\tto failure");
	  fprintf(ff,"\n\t\tend");
      }
      break;

    case RetryUntilSuccessful:

      btnode1 = sl_get_slist_head(btnode->children);

      fprintf(ff,"\n\tto %s_%s", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);

      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tretry := 0;");
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tretry := retry + 1 ;");
      fprintf(ff,"\n\t\t\tif (retry = attempts) then");
      fprintf(ff,"\n\t\t\t\tretry := 0;");
      fprintf(ff,"\n\t\t\t\tto failure");
      fprintf(ff,"\n\t\t\telse");
      fprintf(ff,"\n\t\t\t\tto tick_node");
      fprintf(ff,"\n\t\t\tend");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case Recovery:
      index = 1;
      sl_loop_through_slist(btnode->children, btnode1, BTNode *) {
	fprintf(ff,"\n\t\tif (next_rec = %d) then to %s_%s end;", index, btnode1->id, btnode1->name);
	index++;
      }
      fprintf(ff,"\n\t\tto error");

      if (index != 3) fprintf(stderr, "ERROR: expected exactly two children for a recovery node (got %d %s_%s).", index-1, btnode->id, btnode->name);

      index = 1;
      btnode1 = sl_get_slist_next(btnode->children, NULL);
      btnode2 = sl_get_slist_next(btnode->children, btnode1);

      fprintf(ff,"\n\n\tfrom %s_%s", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling the first node %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode1->id, btnode1->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode1->id, btnode1->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode1->id, btnode1->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tnext_rec := 1;"); /* reset it */
      fprintf(ff,"\n\t\t\tto success");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode1->id, btnode1->name);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' failed its 1st node %s '%s_%s', attempting a recovery with the second one %s '%s_%s'",
		  BTNode_Type_Name[btnode->type], btnode->id,  btnode->name,
		  BTNode_Type_Name[btnode1->type], btnode1->id,  btnode1->name,
		  BTNode_Type_Name[btnode2->type], btnode2->id,  btnode2->name);
      fprintf(ff,"\n\t\t\tnext_rec := 2;"); /* next call will get the second node */
      fprintf(ff,"\n\t\t\tto %s_%s", btnode2->id, btnode2->name);
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode1->id, btnode1->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      fprintf(ff,"\n\n\tfrom %s_%s", btnode2->id, btnode2->name);
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' calling the second node %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode2->type], btnode2->id,  btnode2->name);
      fprintf(ff,"\n\t\tbtnode[%s_%s].caller := caller_%s_%s;", btnode2->id, btnode2->name, btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_tick_passed);
      fprintf(ff,"\n\t\tto %s_%s_done", btnode2->id, btnode2->name);
	
      fprintf(ff,"\n\n\tfrom %s_%s_done", btnode2->id, btnode2->name);
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\t\ton (btnode[%s_%s].caller = None);", btnode2->id, btnode2->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_ticked);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' getting control back from %s '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode2->type], btnode2->id,  btnode2->name);
      fprintf(ff,"\n\t\tif (btnode[%s_%s].rstatus = success) then", btnode2->id, btnode2->name);
      fprintf(ff,"\n\t\t\tnext_rec := 1;"); /* goes back to first node */
      fprintf(ff,"\n\t\t\tto %s_%s", btnode1->id, btnode1->name); /* back to first node */
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = failure) then", btnode2->id, btnode2->name);
      HIPPO_TRACE(4,"\n\t\t","Btnode %s '%s_%s' got a faileure from the second node '%s_%s'", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, BTNode_Type_Name[btnode2->type], btnode2->id,  btnode2->name);
      fprintf(ff,"\n\t\t\tnext_rec := 1;"); /* reset it */
      fprintf(ff,"\n\t\t\tto failure");
      fprintf(ff,"\n\t\telsif (btnode[%s_%s].rstatus = running) then", btnode2->id, btnode2->name);
      fprintf(ff,"\n\t\t\tto running");
      fprintf(ff,"\n\t\telse");
      fprintf(ff,"\n\t\t\tto error");
      fprintf(ff,"\n\t\tend");

      break;

    case AsyncSequence:
    case AsyncFallback:
    case IfThenElse:
    case WhileDoElse:
    case SubTree:
    case UnknownBTNodeType:
    default:
      fprintf(stderr, "ERROR: unsupported BT node type for now: %s\n", BTNode_Type_Name[btnode->type]);
      fprintf(ff,"\n\tto error");
      break;
    }

    if (btnode->type != AlwaysFailure && btnode->type != ForceFailure && btnode->type != KeepRunningUntilFailure) {
      fprintf(ff,"\n\n\tfrom success");
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(2,"\n\t\t","Btnode %s '%s_%s'%sreturns success.", BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
      fprintf(ff,"\n\t\tbtnode[%s_%s].rstatus := success;", btnode->id, btnode->name);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_success);
      fprintf(ff,"\n\t\tto done");  
    }
    
    if (btnode->type != SetSV && btnode->type != AlwaysSuccess && btnode->type != ForceSuccess &&
	! (btnode->type == Eval && (ASSIGN_P(btnode->expr)))) {
      fprintf(ff,"\n\n\tfrom failure");
      fprintf(ff,"\n\t\twait [0,0];");
      if (btnode->type != Condition && bt2f_option[HIPPO]) {
	fprintf(ff,"\n\t\tif (report_halted) then");
	HIPPO_TRACE(2,"\n\t\t\t","Btnode %s '%s_%s'%sreturns halted failure.", BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
	GRAPHIC_TRACE("\n\t\t\t", btnode, ns_halted);
	fprintf(ff,"\n\t\t\tnull");
	fprintf(ff,"\n\t\telse");
      }
      HIPPO_TRACE(2,"\n\t\t\t","Btnode %s '%s_%s'%sreturns failure.", BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
      GRAPHIC_TRACE("\n\t\t\t", btnode, ns_failure);
      if (btnode->type != Condition && bt2f_option[HIPPO]) {
	fprintf(ff,"\n\t\t\tnull");
	fprintf(ff,"\n\t\tend;");
      }
      fprintf(ff,"\n\t\tbtnode[%s_%s].rstatus := failure;", btnode->id, btnode->name);
      fprintf(ff,"\n\t\tto done");
      
      if (btnode->type != Condition) {

	fprintf(ff,"\n\n\tfrom halted");
	fprintf(ff,"\n\t\twait [0,0];");
	HIPPO_TRACE(2,"\n\t\t","Btnode %s '%s_%s'%shas been halted.", BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
	if (btnode->type == Sequence || btnode->type == ReactiveSequence ||
	    btnode->type == RoundRobin || btnode->type == SequenceWithMemory) {
	  fprintf(ff,"\n\t\tnext_seq := 1;");
	}
	if (btnode->type == Fallback || btnode->type == ReactiveFallback) {
	  fprintf(ff,"\n\t\tnext_fb := 1;");
	}
	if (btnode->type == Recovery) {
	  fprintf(ff,"\n\t\tnext_rec := 1;");
	}
	if (btnode->type == PipelineSequence) {
	  fprintf(ff,"\n\t\tlast_succ := 1;");
	}

	if (btnode->type == Parallel || btnode->type == ParallelAll || btnode->type == ReactiveSequence ||
	    btnode->type == RoundRobin || btnode->type == PipelineSequence) {
	  fprintf(ff,"\n\t\tif (halt_called_by_me) then");
	  fprintf(ff,"\n\t\t\tif (force_success) then");
	  fprintf(ff,"\n\t\t\t\tto success");
	  fprintf(ff,"\n\t\t\telse");
	  fprintf(ff,"\n\t\t\t\tto failure");
	  fprintf(ff,"\n\t\t\tend");
	  fprintf(ff,"\n\t\telse");
	  if (bt2f_option[HIPPO])
	    fprintf(ff,"\n\t\t\treport_halted := true;");
	  fprintf(ff,"\n\t\t\tto failure");
	  fprintf(ff,"\n\t\tend");
	} else {
	  if (bt2f_option[HIPPO])
	    fprintf(ff,"\n\t\treport_halted := true;");
	  fprintf(ff,"\n\t\tto failure");
	}
      }
    }

    
    if (btnode->type != Condition && btnode->type != Eval && btnode->type != SetSV &&
	(btnode->type != Action || btnode->PossibleRet == SFR || btnode->PossibleRet == SR ||
	 btnode->PossibleRet == FR || btnode->PossibleRet == R)) {
      fprintf(ff,"\n\n\tfrom running");
      fprintf(ff,"\n\t\twait [0,0];");
      HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s'%sreturns running.", BTNode_Type_Name[btnode->type], btnode->id, btnode->name, arg_string);
      GRAPHIC_TRACE("\n\t\t", btnode, ns_running);
      fprintf(ff,"\n\t\tbtnode[%s_%s].rstatus := running;", btnode->id, btnode->name);
      if (sl_in_slist(bt_list,btnode)) { /* if a top level node, and running, keep runnning */
	fprintf(ff,"\n\t\tto tick_node");
      } else {
	fprintf(ff,"\n\t\tto done");
      }
    }

    fprintf(ff,"\n\n\tfrom done");
    fprintf(ff,"\n\t\twait [0,0];");
    if (btnode->type == Parallel || btnode->type == ParallelAll || btnode->type == ReactiveSequence ||
	btnode->type == RoundRobin || btnode->type == PipelineSequence) {
      fprintf(ff,"\n\t\thalt_called_by_me := false;");
      fprintf(ff,"\n\t\tforce_success := false;");
    }
    HIPPO_TRACE(3,"\n\t\t","Btnode %s '%s_%s' returning control to caller and back to 'ether'", BTNode_Type_Name[btnode->type], btnode->id, btnode->name);
    fprintf(ff,"\n\t\tbtnode[%s_%s].caller := None;", btnode->id, btnode->name);
    fprintf(ff,"\n\tto ether");

    if (!sl_in_slist(bt_list,btnode)) { /* if not a top level node, we loop */
      fprintf(ff,"\n\n\tfrom ether");
      fprintf(ff,"\n\t\twait [0,0];");
      fprintf(ff,"\n\tto start_");
    }
  }
  return TRUE;
}

PBoolean btcomponent_to_fcr(FILE *ff, Slist *svl //,
			    // Slist* skilll //, Slist* eventl, Slist* resl
			    )
{
  Sv *sv;
  /* Res *res; */
  BTNode *btnode;
  /* Event *event; */
  PBoolean first;
  /* BranchProcess *bp;     */

  const BTNode *top_btree = sl_get_slist_tail(bt_list);
  int sv_automata = 0;

  fprintf(ff,"\n\ncomponent %s is", top_btree->name);

  fprintf(ff,"\n\tvar");

  fprintf(ff,"\n\t\tbtnode: btnode_array := [");
  first = TRUE;
  //  int index = sl_slist_length (btnode_list);
  sl_loop_through_slist(btnode_list, btnode, BTNode *) {
    if (sl_in_slist(bt_list,btnode))
      fprintf(ff,"%s\n\t\t\t{caller = caller_%s_%s", (first?"":", "), btnode->id, btnode->name); /* the last one we consider to be the main. */
    else
      fprintf(ff,"%s\n\t\t\t{caller = None", (first?"":", "));
    fprintf(ff,", ArgIndex = 0, rstatus = no_ret_status}");
    first = FALSE;
  }
  fprintf(ff,"]");

  if (! sl_slist_empty(sv_list)) {
  fprintf(ff,",\n\t\t// **** state variables ****");
  sl_loop_through_slist(sv_list, sv, Sv *) {
    if (sv->init) {
      fprintf(ff,",\n\t\t%s: sv_%s := %s", sv->name, sv->name, sv->init);
      sv_automata++;
  } else {
      fprintf(ff,",\n\t\t%s: sv_%s := %d", sv->name, sv->name, sv->initi);
    }
  }
  }
  /* if (! sl_slist_empty(resl)) { */
  /*   fprintf(ff,",\n\t\t// **** resources ****"); */
  /*   sl_loop_through_slist(resl, res, Res *) { */

  /*     fprintf(ff,",\n\t\t%s: res_%s := %d", res->name, res->name, res->init); */
  /*   } */
  /* } */

  /* fprintf(ff,"\n\tport"); */

  /* fprintf(ff,"\n\t\t// **** event ports ****"); */

  /* first = TRUE; */
  /* sl_loop_through_slist(eventl, event, Event *) { */

  /*   if (!first) fprintf(ff,","); */
  /*   else first = FALSE; */

  /*   fprintf(ff,"\n\t\t%s: sync", event->name); */
  /* } */
  /* fprintf(ff,",\n\t\t// **** interrupt ports ****"); */
  /* sl_loop_through_slist(btnodel, btnode, BTNode *) { */
  /*   if (btnode->interrupt) { */

  /*     if (!first) fprintf(ff,","); */
  /*     else first = FALSE; */

  /*     fprintf(ff,"\n\t\tinterrupt_%s: sync", btnode->name); */
  /*   } */
  /* } */


  fprintf(ff,"\n\npar * in");
  first = TRUE;

  fprintf(ff,"\n// **** state variables processes ****");

  sl_loop_through_slist(sv_list, sv, Sv *) {

    if (!sv->init) continue;

    if (!first) fprintf(ff,"\n|| ");
    else {
      first = FALSE;
      fprintf(ff,"\n   ");
    }
    fprintf(ff,"sv_%s_automata(&%s)", sv->name, sv->name);
  }

  /* fprintf(ff,"\n// **** event processes ****"); */
  /* sl_loop_through_slist(eventl, event, Event *) { */
  /*   PBoolean first_arg; */

  /*   fprintf(ff,"\n|| event_%s_automata\n\t[%s]\n\t(", event->name,  event->name); */

  /*   first_arg = TRUE; */
  /*   sl_loop_through_slist(svl, sv, Sv *) { */

  /*     if (!first_arg) fprintf(ff,", "); */
  /*     else first_arg = FALSE; */
  /*     fprintf(ff,"&%s", sv->name); */
  /*   } */
  /*   fprintf(ff,")"); */
  /* } */

  fprintf(ff,"\n// **** btnode processes ****");

  sl_loop_through_slist(btnode_list, btnode, BTNode *) {
    //    PBoolean first_arg;

    /* if (btnode->interrupt)  */
    /*   fprintf(ff,"\n%s btnode_%s\n\t[interrupt_%s]\n\t(", (first?"  ","||"),btnode->name, btnode->name); */
    /* else */
    fprintf(ff,"\n%s btnode_%s_%s(", (first?"  ":"||"), btnode->id, btnode->name);

    first = FALSE;

    fprintf(ff,"&btnode");

    if (btnode->type == Action || btnode->type == Condition || btnode->type == Eval || btnode->type == SetSV) {
      sl_loop_through_slist(svl, sv, Sv *) {
	fprintf(ff,", &%s", sv->name);
      }
    }
    /* sl_loop_through_slist(resl, res, Res *) { */
    /*   fprintf(ff,", &%s", res->name); */
    /* } */
    fprintf(ff,")");
  }

  /* fprintf(ff,"\n// **** btnode watchdog processes ****"); */
  /* sl_loop_through_slist(btnodel, btnode, BTNode *) { */
  /*   if (btnode->btnode->body && btnode->time_interval) { */
  /*     fprintf(ff,"\n|| btnode_%s_watchdog(&btnode) ", btnode->name); */
  /*   }     */
  /* } */

  /* fprintf(ff,"\n// **** btnode parallel branch processes ****"); */
  /* sl_loop_through_slist(btnodel, btnode, BTNode *) { */
  /*   sl_loop_through_slist(btnode->dbp, bp, BranchProcess *) { */

  /*     if (btnode->interrupt)  */
  /* 	fprintf(ff,"\n|| btnode_%s_branch_%d_%d\n\t[interrupt_%s]\n\t(", btnode->name, bp->level, bp->index, btnode->name); */
  /*     else */
  /* 	fprintf(ff,"\n|| btnode_%s_branch_%d_%d\n\t(", btnode->name, bp->level, bp->index); */

  /*     fprintf(ff,"&btnode"); */

  /*     sl_loop_through_slist(svl, sv, Sv *) { */
  /* 	fprintf(ff,", &%s", sv->name); */
  /*     } */
  /*     sl_loop_through_slist(resl, res, Res *) { */
  /* 	fprintf(ff,", &%s", res->name); */
  /*     } */
  /*     fprintf(ff,")"); */
  /*   }     */
  /* } */

  /* fprintf(ff,"\n// **** basic environment processes ****"); */
  /* fprintf(ff,"\n|| basic_environment_%s\n\t[", current_btnodeset->name); */

  /* first = TRUE; */

  /* sl_loop_through_slist(eventl, event, Event *) { */
  /*   if (!first) fprintf(ff,", "); */
  /*   else first = FALSE; */
  /*   fprintf(ff,"%s", event->name); */
  /* } */
  /* //  fprintf(ff,"\n\t"); */

  /* sl_loop_through_slist(btnodel, btnode, BTNode *) { */
  /*   if (btnode->interrupt) { */
  /*     if (!first) fprintf(ff,", "); */
  /*     else first = FALSE; */
  /*     fprintf(ff,"interrupt_%s", btnode->name); */
  /*   } */
  /* } */

  /* fprintf(ff,"]\n\t("); */

  /* fprintf(ff,"&btnode"); */

  /* sl_loop_through_slist(btnodel, btnode, BTNode *) { */
  /*   fprintf(ff,", &run_btnode_%s", btnode->name); */
  /* } */

  /* sl_loop_through_slist(svl, sv, Sv *) { */
  /*   fprintf(ff,", &%s", sv->name); */
  /* } */

  /* fprintf(ff,")"); */

  fprintf(ff,"\n// **** end of the component ****");
  fprintf(ff,"\n\nend");

  fprintf(ff,"\n\n\n%s\n", top_btree->name);

  if (!bt2f_option[HIPPO]) {

    fprintf(ff,"\n/* Some properties to model check */");

    int i = sv_automata;

    /* sl_loop_through_slist(svl, sv, Sv *) { */
    /*   if (sv->init) i++;	/\* count the automata SV *\/ */
    /* } */

    sl_loop_through_slist(btnode_list, btnode, BTNode *) {
	i++;

	fprintf(ff,"\n\n/* Can the %s btnode properly execute? */",  btnode->name);
	fprintf(ff,"\nproperty %s_%s_%s_undoable is absent %s/%d/state ether",
		BTNode_Type_Name[btnode->type], btnode->id,  btnode->name,  top_btree->name, i);
	fprintf(ff,"\nprove %s_%s_%s_undoable", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);

	fprintf(ff,"\n\n/* Can the %s btnode success? */",  btnode->name);
	fprintf(ff,"\nproperty %s_%s_%s_cannot_succeed is absent %s/%d/value (btnode[%s_%s].rstatus=success)",
		BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, top_btree->name, i, btnode->id, btnode->name);
	fprintf(ff,"\nprove %s_%s_%s_cannot_succeed", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);

	fprintf(ff,"\n\n/* Can the %s btnode return failure? */",  btnode->name);
	fprintf(ff,"\nproperty %s_%s_%s_cannot_fail is absent  %s/%d/value (btnode[%s_%s].rstatus=failure)",
		BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, top_btree->name, i, btnode->id, btnode->name);
	fprintf(ff,"\nprove %s_%s_%s_cannot_fail", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);

	if (btnode->type != Condition) {
	  fprintf(ff,"\n\n/* Can the %s btnode be halted? */",  btnode->name);
	  fprintf(ff,"\nproperty %s_%s_%s_cannot_be_halted is absent %s/%d/state halted",
		  BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, top_btree->name, i);
	  fprintf(ff,"\nprove %s_%s_%s_cannot_be_halted", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);
	}
	
	if (btnode->type != Parallel && btnode->type != ParallelAll && btnode->type != Condition && 
	    btnode->type != Eval && btnode->type != SetSV &&
	    (btnode->type != Action || btnode->PossibleRet == SFR || btnode->PossibleRet == SR ||
	     btnode->PossibleRet == FR || btnode->PossibleRet == R)) {
	  fprintf(ff,"\n\n/* Can the %s btnode return running? */",  btnode->name);
	  fprintf(ff,"\nproperty %s_%s_%s_cannot_return_running is absent %s/%d/state running",
		  BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, top_btree->name, i);
	  fprintf(ff,"\nprove %s_%s_%s_cannot_return_running", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);
	}
	
	if (btnode->type != Parallel && btnode->type != ParallelAll && btnode->type != SetSV && btnode->type != Eval &&
	    btnode->type != Eval && btnode->type != Condition && btnode->type != Condition && btnode->type != Action) {
	  fprintf(ff,"\n\n/* Can the %s btnode go in error? */",  btnode->name);
	  fprintf(ff,"\nproperty %s_%s_%s_cannot_error is absent %s/%d/state error",
		  BTNode_Type_Name[btnode->type], btnode->id,  btnode->name, top_btree->name, i);
	  fprintf(ff,"\nprove %s_%s_%s_cannot_error", BTNode_Type_Name[btnode->type], btnode->id,  btnode->name);
	}
    }
  }

  fprintf(ff,"\n/*End of File */\n");
  return TRUE;
}

PBoolean write_fcr(PString file_name)
{
  FILE *ff;			/* Fiacre file */
  FILE *hf = NULL;			/* h file */
  FILE *hargf = NULL;			/* h file */
  /* Slist *res_list = current_btnodeset->res_list; */
  /* Slist *sv_list = current_btnodeset->sv_list; */
  /* Slist *event_list = current_btnodeset->event_list; */

  // Slist *bt_list = sl_copy_slist(current_btnodeset->bt_list);

  hippo_trace_list = sl_make_slist();
  hippo_trace_index = 0;
  hippo_user_print_list = sl_make_slist();
  hippo_user_print_index = 0;

  if (!(ff = (FILE *) fopen(file_name, "w"))) {
    fprintf(stderr, LG_STR("ERROR: write_fcr: failed fdopen %s\n",	
		   "ERREUR: save_database: echec fdopen %s\n"), file_name);
    return FALSE;
  }

  //  btnode_list = sl_make_slist();

  //  add_btnode_to_btnodelist(bt_list);

  if (bt2f_option[HIPPO]) {
    PString suffix;
    Sprinter *sp = make_sprinter(0);

    NEWSTR(file_name, suffix);

    char *index = strrchr(suffix,'.');
    index[0] = '\0';


    SPRINT(sp,2+strlen(suffix),sprintf(f,"%s.h", suffix));

    if (!(hf = (FILE *) fopen(sprinter_string(sp), "w"))) {
      fprintf(stderr, "ERROR: write_h: failed fdopen %s\n", sprinter_string(sp));
      return FALSE;
    }

    reset_sprinter(sp);

    SPRINT(sp,15+strlen(suffix),sprintf(f,"%s-action-arg.h", suffix));

    if (!(hargf = (FILE *) fopen(sprinter_string(sp), "w"))) {
      fprintf(stderr, "ERROR: write_harg: failed fdopen %s\n", sprinter_string(sp));
      return FALSE;
    }
    FREE(suffix);
  }
  fprintf(ff,COPYRIGHT_FILE_HEADER);
  fprintf(ff,
	  "\n\
/*\n\
 * This Behavior Tree FIACRE model has been automatically generated by bt2f %s.\n\
 * DO NOT EDIT.\n\
 */\n", package_version);

  if (bt2f_option[HIPPO])
    fprintf(ff,
	    "\n\
/*\n								\
 * This Behavior Tree FIACRE model is for the HIPPO engine.\n	\
 */\n");
  else
    fprintf(ff,
	    "\n\
/*\n								\
 * This Behavior Tree FIACRE model is for verification with the TINA tools.\n	\
 */\n");
  
  if (! btnode_list_to_global_declaration(ff, btnode_list)) return FALSE;
  if (! sv_list_to_fcr(ff, sv_list)) return FALSE;
  if (! btnode_list_to_fiacre_declaration(ff, btnode_list)) return FALSE;

  // if (! res_list_to_fcr(ff, res_list)) return FALSE;
  // if (! event_list_to_fcr(ff, sv_list, event_list)) return FALSE;

  if (! btnode_list_to_fcr(ff, sv_list, // res_list,
		       btnode_list)) return FALSE;

  // if (! basic_environment_to_fcr(ff, sv_list, bt_list, event_list)) return FALSE;

  if (! btcomponent_to_fcr(ff, sv_list //,
			   // btnode_list // , event_list, res_list
			   )) return FALSE;

  if (bt2f_option[HIPPO]) {
    output_trace_string(hf);
    fclose(hf);
    output_arg_string(hargf);
    fclose(hargf);
  }
  fclose(ff);

  //  FREE_SLIST(bt_list);
  return TRUE;
}


