/*                               -*- Mode: C -*- 
 * bt2f-init.c -- 
 * 
 * Copyright (c) 1991-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#include <fcntl.h>
#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef HAS_SYS_SIGNAL
#include <sys/signal.h>
#else
#include <signal.h>
#endif

#include "shashPack_f.h"

#include "constant.h"
#include "version.h"
#include "macro.h"

#include "bt2f-main.h"
#include "bt2f-type.h"
#include "bt-structure.h"
#include "parser-funct.h"
#include "bt2f.h"
#include "top-lev.h"
#include "bt2f-error.h"
#include "bt2f-sprint.h"
#include "bt2f-lex.h"

#include "yy-funct_f.h"
#include "help_f.h"
#include "parser-funct_f.h"
#include "bt2f-pred-func_f.h"
#include "bt2f-init_f.h"
#include "bt2f-socket_f.h"
#include "bt2f-type_f.h"
#include "bt2f-util_f.h"
#include "bt2f_f.h"
#include "top-lev_f.h"

#include "bt2f-error_f.h"
#include "bt2f-sprint_f.h"

void install_quit_handler(void);

//#define BT2F_HELP_MESSAGE ""

#define BT2F_ARG_ERR_MESSAGE LG_STR(\
"Usage: %s [-x include-filename]* [-c command]* [-l upper|lower|none ] [-L en|fr]\n\
\t[-I size-ID-hash-table] [-P size-pred-hash-table]\n\
\t[-D size-database-hash-table]\n\
\t[-n] client-name\n",\
"Utilisation: %s [-x include-fichier]* [-c commande]* [-l upper|lower|none ] [-L en|fr]\n\
\t[-I taille-table-ID] [-P taille-table-predicat]\n\
\t[-D taille-table-database]\n\
\t[-d bt2f-data-path] [-n] nom-client\n")

PBoolean register_to_server = TRUE;
PBoolean register_to_mp = TRUE;

PBoolean stop_checking_the_stdin = TRUE;

PBoolean dumped_code = FALSE;

extern char *server_hostname;
extern int server_port;

Op_Instance_List previous_soak;

void wrap_up(void)
{
#if defined(HAVE_SETITIMER) && defined(WANT_TRIGGERED_IO)
     kill_condition_timer();
#endif
     //     end_kernel_hook();
     free_bt2f(current_bt2f);
     current_bt2f = NULL;

     free_pred_func_hash();
     // free_global_var_list();
     free_id_hash();
     //     free_type_hash();

     close(0);
     /* if (register_to_server)  */
     /* 	  if (CLOSE_SOCK(ps_socket) < 0) */
     /* 	       perror("wrap_up: ps_socket: close/closesocket"); */
     /* if (register_to_mp)  */
     /* 	  disconnect_from_the_mp(); */

     FREE_SLIST(empty_list);
     /* FREE_OBJECT(goal_satisfied_in_db_op_ptr); */
     /* FREE_OBJECT(goal_for_intention_op_ptr); */
     /* FREE_OBJECT(goal_waiting_op_ptr); */

     free_dynamic_slist_list();
#ifdef BT2F_DEBUG
     pmstats();
#endif
}

Slist *list_buffered_command = NULL;

void save_command_in_buffer (PString command)
{
     PString buffered_command;
     int total_length;

     if (list_buffered_command == NULL) {
	  list_buffered_command = sl_make_slist();
     }
	
     total_length =  strlen(command) + 1;
     buffered_command = (char *) MALLOC ( total_length);
     sprintf (buffered_command, "%s", command);
     sl_add_to_tail(list_buffered_command, buffered_command);
}

extern PBoolean use_dialog_error;
//extern PBoolean flushing_xt_events;


int bt2f_yyparse_caller(void)
{
     closed_file = FALSE;
     yy_begin_COMMAND();
     return bt2f_yyparse();
}

PBoolean parse_one_or_more(void) /* return TRUE if the stop_checking_the_stdin alarm has been raised */
{
     if (parse_source == PS_SOCKET) stop_parsing_from_socket = FALSE;

     bt2f_yyparse_caller();		/* call once */

     if (quit ||  (feof(stdin) != 0)) {
	  wrap_up();
	  socket_cleanup_and_exit(0); /* This is a regular and nominal exit,  */
     }     
     
     while (((parse_source == PS_SOCKET) &&
	     !stop_parsing_from_socket && /* The socket is emtpy */
	     !stop_checking_the_stdin) || /* enough commands have come now...  */
	    i_have_the_stdin || (parser_index > 0)) {
	  if (quit || (feof(stdin) && (parser_index == 0))) {
	       wrap_up();
	       socket_cleanup_and_exit(0); /* This is a regular and nominal exit,  */
	  }     
	  print_bt2f_prompt();
	  bt2f_yyparse_caller();
     }

     /* if ( use_dialog_error && (parser_message != NULL)) */
     /* 	  report_syntax_error(""); */

     if (list_buffered_command  != NULL) { /* A command has been saved in buffer, parse it */
	       Slist *tmp_list = list_buffered_command;
	       PString buffered_command;
	       
	       list_buffered_command = NULL; /* to do not loop in recursive call */
	       sl_loop_through_slist(tmp_list, buffered_command, PString) {
		    send_command_to_parser(buffered_command);
		    FREE(buffered_command);
	       }
	       FREE_SLIST(tmp_list);
     }
     
     return stop_checking_the_stdin;
}

void send_command_to_parser(PString s)
{
	  Parse_Source_Type string_saved_parse_source;
	  int string_saved_line_number;
     
	  string_saved_parse_source = parse_source;
	  string_saved_line_number = string_line_number;
	  parse_source = PS_STRING;  
	  string_line_number = 1;
	  bt2f_yy_scan_string(s);	   
 	  yy_begin_COMMAND();		/* On demarre le lex parser dans cet etat */
	  safe_parse_one_or_more();
	  yy_delete_buffer_no_arg();
	  parse_source = string_saved_parse_source;
	  string_line_number = string_saved_line_number;
}

void report_meta_option_array(void)
{
     PBoolean one_is_on = FALSE;

     fprintf(stdout, LG_STR("The following meta-fact are posted:",
			    "Les mta faits suivant seront posts:"));

     if (meta_option[META_LEVEL] && meta_option[SOAK_MF]) {
	  fprintf(stdout, LG_STR(" SOAK",
				 " SOAK"));
	  one_is_on = TRUE;
     }
     if (meta_option[META_LEVEL] && meta_option[FACT_INV]){
	  fprintf(stdout, LG_STR(" FACT-INVOKED-OPS",
				 " FACT-INVOKED-OPS"));
	  one_is_on = TRUE;
     }
     if (meta_option[META_LEVEL] && meta_option[GOAL_INV]){
	  fprintf(stdout, LG_STR(" GOAL-INVOKED-OPS",
				 " GOAL-INVOKED-OPS"));
	  one_is_on = TRUE;
     }
     if (meta_option[META_LEVEL] && meta_option[APP_OPS_FACT]){
	  fprintf(stdout, LG_STR(" APPLICABLE-OPS-FACT",
				 " APPLICABLE-OPS-FACT"));
	  one_is_on = TRUE;
     }
     if (meta_option[META_LEVEL] && meta_option[APP_OPS_GOAL]){
	  fprintf(stdout, LG_STR(" APPLICABLE-OPS-GOAL",
				 " APPLICABLE-OPS-GOAL"));
	  one_is_on = TRUE;
     }
     if (one_is_on)
	  fprintf(stdout, "\n");
     else
	  fprintf(stdout, "none.\n");
}

void initialize_option_array()
{
#ifdef BT2F_PROFILING
     profiling_option[PROFILING] = TRUE;
     profiling_option[PROF_DB] = TRUE;
     profiling_option[PROF_ACTION_OP] = TRUE;
     profiling_option[PROF_OP] = TRUE;
     profiling_option[PROF_EP] = TRUE;
     profiling_option[PROF_EF] = TRUE;
     profiling_option[PROF_AC] = TRUE;
#endif
     bt2f_option[HIPPO] = FALSE;
     bt2f_option[TINA_COMPACT] = FALSE;

     meta_option[META_LEVEL] = TRUE;
     meta_option[SOAK_MF] = TRUE;
     meta_option[FACT_INV] = TRUE;
     meta_option[GOAL_INV] = TRUE;
     meta_option[APP_OPS_FACT] = FALSE;
     meta_option[APP_OPS_GOAL] = FALSE;

     compiler_option[CHECK_ACTION] = TRUE;
     compiler_option[CHECK_PFR] = TRUE;
     compiler_option[CHECK_SYMBOL] = TRUE;

     run_option[EVAL_ON_POST] = TRUE;
     run_option[PAR_POST] = TRUE;
     run_option[PAR_INTEND] = TRUE;
     run_option[PAR_EXEC] = TRUE;

     user_trace[USER_TRACE] = FALSE;
     user_trace[UT_POST_GOAL] = FALSE;
     user_trace[UT_POST_FACT] = FALSE;
     user_trace[UT_RECEIVE_MESSAGE] = FALSE;
     user_trace[UT_SEND_MESSAGE] = FALSE;
/*      user_trace[UT_DATABASE] = FALSE;
     user_trace[UT_INTENTION_FAILURE] = FALSE; */
     user_trace[UT_SUC_FAIL] = FALSE;
     user_trace[UT_INTEND] = FALSE;
     user_trace[UT_RELEVANT_OP] = FALSE;
     user_trace[UT_SOAK] = FALSE;
     user_trace[UT_OP] = FALSE;
     user_trace[UT_THREADING] = FALSE;


     //     report_meta_option_array();
}

void initialize_debug_trace_array()
{
     debug_trace[INTENTION_FAILURE] = TRUE;
     debug_trace[INT_FAIL_BIND] = FALSE;
     debug_trace[SUC_FAIL] = FALSE;
     debug_trace[DATABASE] = FALSE;
     debug_trace[DATABASE_KEY] = FALSE;
     debug_trace[DATABASE_FRAME] = FALSE;
     debug_trace[CONCLUDE_FROM_PARSER] = TRUE;
     debug_trace[POST_GOAL] = FALSE;
     debug_trace[POST_FACT] = FALSE;
     debug_trace[RECEIVE_MESSAGE] = TRUE;
     debug_trace[SEND_MESSAGE] = FALSE;
     debug_trace[INTEND] = FALSE;
     debug_trace[OP_COMPILER] = FALSE;
     debug_trace[RELEVANT_OP] = FALSE;
     debug_trace[SOAK] = FALSE;
     debug_trace[GRAPHIC_OP] = TRUE;
     debug_trace[TEXT_OP] = TRUE;
     debug_trace[GRAPHIC_INTEND] = TRUE;
     debug_trace[THREADING] = FALSE;
}

void initialize_cst()
{
     empty_list = sl_make_slist();	/* create the empty list (constant please...) */
}

void bt2f_init_env(void)
{
  // bt2f_data_path = getenv("BT2F_DATA_PATH");
}
Slist *initial_commands = NULL;

void run_initial_commands(void) 
{
     if (initial_commands) {
	  char *command = (char *) sl_get_from_head(initial_commands);
	  while (command != NULL) {
	       send_command_to_parser(command); 
	       FREE (command);
	       command = (char *) sl_get_from_head(initial_commands);
	  }
	  FREE_SLIST(initial_commands);
     }
}

void add_initial_command(char *command_arg)
{
     char * command= (char *)MALLOC(strlen(command_arg) + 1 + 1);

     sprintf(command,"%s\n",command_arg);
     SAFE_SL_ADD_TO_TAIL(initial_commands, command);
}

void add_initial_include_command(char *include_filename)
{
     Sprinter *sp = make_sprinter(0);
     SPRINT(sp, 16 + strlen(include_filename),
	    sprintf(f,"include \"%s\"\n", include_filename)); 
     add_initial_command(SPRINTER_STRING(sp));

     free_sprinter(sp);
}

char *bt2f_init_arg(int argc,char **argv)
{
     char *bt2f_name = NULL, *lang_str = NULL;	/* Just to avoid GCC Warning */
     //     char *data_path = NULL;
     char *id_case_option = NULL;
     //     char *mp_hostn_g, *mp_hostname = NULL;
     //     int mp_port;
     //     char *server_hostn_g, *server_hostname = NULL;
     //     int server_port;


     int c, getoptflg = 0;
     int /* sname_flg = 0, snumber_flg = 0, mpname_flg = 0, mpnumber_flg = 0, */ bt2fname_flg = 0;
     int /* quiet_flg = 0, dname_flg = 0, */ alone_flg = 0, alone_mp_flg = 0, english_flg = 0, lci_flg = 0;
     int /* database_flg = 0,*/ pred_func_flg = 0, id_flg = 0, lang_flg = 0/* , tcl_flg = 0 */;
     int /* database_hash_size,*/ pred_func_hash_size, id_hash_size;

     extern int optind;
     extern char *optarg;
     //     int maxlength = MAX_HOST_NAME * sizeof(char);

     while ((c = getopt(argc, argv, 
			"c:d:s:x:i:m:j:n:D:P:I:l:L:htpqaA"
	  )) != EOF)
	  switch (c)
	  {
	  case 'x': 
	       add_initial_include_command(optarg);
	       break;
	  case 'c': 
	       add_initial_command(optarg);
	       break;
	  /* case 'd': */
	  /*      dname_flg++; */
	  /*      data_path = optarg; */
	  /*      break; */
	  /* case 's': */
	  /*      sname_flg++; */
	  /*      NEWSTR(optarg,server_hostname); */
	  /*      break; */
	  /* case 'i': */
	  /*      snumber_flg++; */
	  /*      if (!sscanf (optarg,"%d",&server_port )) */
	  /* 	    getoptflg++; */
	  /*      break; */
	  /* case 'm': */
	  /*      mpname_flg++; */
	  /*      NEWSTR(optarg, mp_hostname); */
	  /*      break; */
	  /* case 'j': */
	  /*      mpnumber_flg++; */
	  /*      if (!sscanf (optarg,"%d",&mp_port )) */
	  /* 	    getoptflg++; */
	  /*      break; */
	  case 'n':
	       bt2fname_flg++;
	       bt2f_name = optarg;
	       break;
	  /* case 't': */
	  /*      tcl_flg++; */
	  /*      break; */
	  case 'a':
	       alone_flg++;
	       break;
	  case 'A':
	       alone_mp_flg++;
	       break;
	  case 'p':
	       english_flg++;
	       break;
	  case 'l':
	       lci_flg++;
	       id_case_option = optarg;
	       break;
	  case 'L':
	       lang_flg++;
	       lang_str = optarg;
	       break;
	  /* case 'q': */
	  /*      quiet_flg++; */
	  /*      break; */
	  /* case 'D': */
	  /*      database_flg++; */
	  /*      if (!sscanf (optarg,"%d",&database_hash_size )) */
	  /* 	    getoptflg++; */
	  /*      break; */
	  case 'P':
	       pred_func_flg++;
	       if (!sscanf (optarg,"%d",&pred_func_hash_size ))
		    getoptflg++;
	       break;
	  case 'I':
	       id_flg++;
	       if (!sscanf (optarg,"%d",&id_hash_size ))
		    getoptflg++;
	       break;
	  case 'h':
	  default:
	       getoptflg++;
 
	  }
     if (getoptflg) {
	  fprintf(stderr, BT2F_ARG_ERR_MESSAGE, argv[0]);
	  socket_cleanup_and_exit(0);
     }

     if ( !bt2fname_flg ){
	  if ( argc < (optind + 1)) {
	       fprintf(stderr, LG_STR("ERROR: We need the name of the bt2f client as argument.\n",
				      "ERREUR: Le nom du bt2f client est un argument obligatoire.\n"));
	       socket_cleanup_and_exit (1);
	  } else
	       bt2f_name = argv[optind];
     }

     if (lang_flg) {
	  select_language(lang_str, FALSE);
     }

     if ( lci_flg ) {
	  if (! check_and_set_id_case_option(id_case_option)) {
	       fprintf(stderr, LG_STR("Invalid case option, check the arguments.\n",
				      "Option de minuscule/majuscule invalide, vrifier les arguments.\n"));
	  }
     } else if ((id_case_option = getenv("BT2F_ID_CASE")) != NULL) {
	  if (! check_and_set_id_case_option(id_case_option)) {
	       fprintf(stderr, LG_STR("Invalid case option, check the BT2F_ID_CASE environment variable.\n",
				      "Option de minuscule/majuscule invalide, vrifier la variable d'environement BT2F_ID_CASE.\n"));
	  }
     } 

     if( bt2f_name[0] == '|'){  /* The name is an "idbar" */
	  int i, length = strlen (bt2f_name) -1;
	  for (i =1 ; i<length ; i++) { /* we check if there is no bar inside the name*/
	       if (bt2f_name[i] == '|'){
		    fprintf(stderr, LG_STR("Error in the bt2f-name.\n",
					   "Erreur dans le nom du bt2f.\n"));
		    socket_cleanup_and_exit(1);
	       }
	  }
	  if (bt2f_name[length] != '|'){
	       fprintf(stderr, LG_STR("Error in the bt2f-name.\n",
				      "Erreur dans le nom du bt2f.\n"));
	       socket_cleanup_and_exit(1);
	  }
     } else
	  NORMALIZE_STRING_IN_PLACE(bt2f_name);

     if (pred_func_flg)
	  init_hash_size_pred_func(pred_func_hash_size);
     else
	  init_hash_size_pred_func(0);

     if (id_flg)
	  init_hash_size_id(id_hash_size);
     else
	  init_hash_size_id(0);

     if ( alone_flg ) {
	  fprintf(stderr, LG_STR("This kernel will not register to the BT2F-Server.\n",
				 "Ce noyau ne s'enregistre pas au BT2F-Server.\n"));
	  register_to_server = FALSE;
     }

     if ( alone_mp_flg ) {
	  fprintf(stderr, LG_STR("This kernel will not register to the message passer.\n",
				 "Ce noyau ne s'enregistre pas au message passer.\n"));
	  register_to_mp = FALSE;
     }

     if ( english_flg ) {
	  fprintf(stderr, LG_STR("This kernel will print english temporal operators.\n",
				 "Ce noyau imprime les optateurs temporel en toutes lettres.\n"));
	  parse_and_print_english_operator = TRUE;
     }


     install_quit_handler();

     return(bt2f_name);
}

Bt2f *bt2f_init_main(char *name)
{
     yy_begin_COMMAND();			/* On demarre le lex parser dans cet etat */

     SRANDOM(getpid());		/* we should use the date or something else... */

     initialize_cst();
     initialize_debug_trace_array();
     initialize_option_array();
     make_id_hash();		/* Make the symbol hash table */
     init_id_hash();

     make_pred_func_hash();

#ifdef BT2F_PROFILING
     init_profile_timer();
#endif


     current_bt2f = make_bt2f(name);

     /* In case we are connected to a file, we want to be unbuffered
      * Otherwise we have trouble when we fork because the buffer of 
      * the child stdin is full of the father command...
      */
     setbuf(stdin,NULL);	

     i_have_the_stdin= FALSE;

     bt2f_prompt = name;
     previous_soak = NULL;
#if defined(HAVE_SETITIMER) && defined(WANT_TRIGGERED_IO)
     install_condition_timer_handler_set_arm();
#endif
     
     print_intro();		/* Welcome to the wonderful world of BT2F */

     return current_bt2f;
}

void set_user_trace(User_Trace ut, PBoolean val)
{
     if (install_user_trace) {
	  user_trace[ut]=val;
     } else {
	  fprintf(stderr, LG_STR("WARNING: set_user_trace: this kernel has no trace user hook.\n\
\tNo need to change these traces.\n",
				 "ATTENTION: set_user_trace: ce noyau n'a pas de fonctions de trace utilisateurs.\n\
\tCa ne sert a rien de changer ces traces.\n"));
     }
}

void sprint_expr_in_gmexpr(Expression *expr)
{
     static Sprinter *sp = NULL;
     
     if (expr == NULL)  
	  return;
     if (OR_P(expr) || AND_P(expr) || NOT_P(expr)) {
	  Term *term;
	  sl_loop_through_slist(expr->terms, term, Term *) 
	       sprint_expr_in_gmexpr(TERM_EXPR(term));
     } else {
	  if (! sp)
	       sp = make_sprinter(0);
	  else
	       reset_sprinter(sp);

	  sprint_expr(sp, expr,TRUE);
	  //	  _write_string_to_socket(ps_socket,SPRINTER_STRING(sp));
     }
}


int count_expr_in_gmexpr(Expression *expr)
{
     int i = 0;
     
     if (expr == NULL)  
	  return 0;
     if (OR_P(expr) || AND_P(expr) || NOT_P(expr)) {
	  Term *term;
	  sl_loop_through_slist(expr->terms, term, Term *) 
	       i += count_expr_in_gmexpr(TERM_EXPR(term));
     } else {
	  i = 1;
     }
     return i;
}

void print_help_from_parser(void)
{
     printf("\n\
The following commands are currently available:\n\
\n\
\t add goal|fact\n\
\t\t Add a goal or a fact in the kernel.\n\
\n\
\t conclude expr\n\
\t\t Conclude the expression in the database.\n\
\n\
\t consult gexpr\n\
\t\t Consult the gexpression in the database.\n\
\n\
\t consult relevant op goal|fact\n\
\t\t Return the relevant op(s) for the goal or fact.\n\
\n\
\t consult applicable op goal|fact\n\
\t\t Return the applicable op(s) for the goal or fact.\n\
\n\
\t declare be predicat\n\
\t\t declare the predicat as basic event predicat.\n\
\n\
\t declare cwp predicat\n\
\t\t declare the predicat as closed world predicat (for the current\n\
\t\t BT2F).\n\
\n\
\t declare ff predicate integer\n\
\t\t declare the predicate as functional fact for the number or\n\
\t\t args integer.\n\
\n\
\t declare op predicate predicate\n\
\t\t declare the predicate as op predicate.\n\
\n\
\t declare id\n\
\t\t declare the symbol id, therefore, the OP compiler will not\n\
\t\t complain if it encounters this symbol.\n\
\n\
\t delete expr \n\
\t\t Delete the expression from the fact database.\n\
\n\
\t delete op op-name \n\
\t\t Delete the op op-name from the op database.\n\
\n\
\t delete opf file-name \n\
\t\t Delete from the op database all the ops from op file file-name.\n\
\n\
\t disconnect\n\
\t\t Instruct the connected bt2f to leave the stdin and give it back\n\
\t\t to the BT2F Server. The bt2f client will return in \"run\" mode.\n\
\n\
\t echo gexpr|gtexpr|gmexpr\n\
\t\t Echo the expression to the screen.\n\
\n\
\t empty fact db\n\
\t\t Clear the database.\n\
\n\
\t empty op db\n\
\t\t Clear the OP database.\n\
\n\
\t help|h|?\n\
\t\t Print this help.\n\
\n\
\t include file_name\n\
\t\t Include and execute all the commands in file_name.\n\
\n\
\t list action\n\
\t\t list all the actions.\n\
\n\
\t list all\n\
\t\t list a bunch of information.\n\
\n\
\t list be\n\
\t\t list the predicate declares as basic event.\n\
\n\
\t list cwp\n\
\t\t list the closed world predicate.\n\
\n\
\t list evaluable function\n\
\t\t list the evaluable functions.\n\
\n\
\t list evaluable predicate\n\
\t\t list the evaluable predicates.\n\
\n\
\t list ff\n\
\t\t list the functional fact predicates.\n\
\n\
\t list function\n\
\t\t list the function.\n\
\n\
\t list op_predicate\n\
\t\t list predicate declare as op_predicate.\n\
\n\
\t list opfs\n\
\t\t list the loaded op files.\n\
\n\
\t list op\n\
\t\t list all the ops loaded.\n\
\n\
\t list predicate\n\
\t\t list all the predicate.\n\
\n\
\t load db file_name\n\
\t\t Load all the facts contained in file_name.\n\
\n\
\t load opf op_graph|file_name\n\
\t\t Compile a list of OPs either from\n\
\t\t stdin or from the file file_name.\n\
\n\
\t q|quit|exit|EOF\n\
\t\t quit the program\n\
\n\
\t reset kernel\n\
\t\t reset the kernel.\n\
\n\
\t save db file_name\n\
\t\t save the database in file_name.\n\
\n\
\t send name message\n\
\t\t send the message to the bt2f named \"name\".\n\
\n\
\t set action on|off\n\
\t\t set the compiler flag action on|off.\n\
\n\
\t set eval post on|off\n\
\t\t Turn on or off the current-quote mechanism.\n\
\n\
\t set meta fact op on|off\n\
\t\t set meta fact op on|off.\n\
\n\
\t set meta fact on|off\n\
\t\t set meta fact on|off.\n\
\n\
\t set meta goal op on|off\n\
\t\t set meta goal op on|off.\n\
\n\
\t set meta goal on|off\n\
\t\t set meta goal on|off.\n\
\n\
\t set meta on|off\n\
\t\t Turn on or off the meta_level mechanism. (this will greatly\n\
\t\t increase the performance of the system).\n\
\n\
\t set parallel intend on|off\n\
\t\t set parallel intending on|off.\n\
\n\
\t set parallel intention on|off\n\
\t\t set parallel intention execution on|off.\n\
\n\
\t set parallel post on|off\n\
\t\t set parallel posting on|off.\n\
\n\
\t set bt2f_data_path string\n\
\t\t set the BT2F_DATA_PATH.\n\
\n\
\t set soak on|off\n\
\t\t set soak on|off.\n\
\n\
\t set time_stamping on|off\n\
\t\t set time stamping on|off.\n\
\n\
\t show copyright\n\
\t\t Print the copyright notice\n\
\n\
\t show db\n\
\t\t show the database contents.\n\
\n\
\t show intention\n\
\t\t show the intention graph.\n\
\n\
\t show op op_name\n\
\t\t display the op named op_name.\n\
\n\
\t show bt2f_data_path\n\
\t\t show BT2F_DATA_PATH value.\n\
\n\
\t show version\n\
\t\t show version.\n\
\n\
\t stat db\n\
\t\t Print Statistics on the database hashtables.\n\
\n\
\t stat id\n\
\t\t Print Statistics on the ID hashtable.\n\
\n\
\t stat all\n\
\t\t Print Statistics on all the hashtables.\n\
\n\
\t trace all on|off\n\
\t\t turn on or off a bunch of trace.\n\
\n\
\t trace applicable op on|off\n\
\t\t trace applicable op on|off.\n\
\n\
\t trace db frame on|off\n\
\t\t trace the frame in database operation.\n\
\n\
\t trace db on|off\n\
\t\t trace database operations.\n\
\n\
\t trace fact on|off\n\
\t\t trace fact posting.\n\
\n\
\t trace goal on|off\n\
\t\t trace goal posting.\n\
\n\
\t trace graphic on|off\n\
\t\t set the graphic trace for op on|off.\n\
\n\
\t trace graphic op op_name on|off\n\
\t\t set the graphic trace for op_name.\n\
\n\
\t trace graphic opf file_name on|off\n\
\t\t set the graphic trace for all the op in opf file_name.\n\
\n\
\t trace intend on|off\n\
\t\t trace intending on|off.\n\
\n\
\t trace load op on|off\n\
\t\t trace op conpilationon|off.\n\
\n\
\t trace receive on|off\n\
\t\t trace message reception on|off.\n\
\n\
\t trace relevant op on|off\n\
\t\t trace relevant op lookup on|off.\n\
\n\
\t trace send on|off\n\
\t\t trace message sending on|off.\n\
\n\
\t trace suc_fail on|off\n\
\t\t trace op success failure on|off.\n\
\n\
\t trace text on|off\n\
\t\t set op text trace on|off.\n\
\n\
\t trace text op op_name\n\
\t\t trace text op op_name.\n\
\n\
\t trace text opf file_name\n\
\t\t trace text opf file_name.\n\
\n\
\t trace thread on|off\n\
\t\t trace threading (split/join) on|off.\n\
\n\
\t undeclare be predicat\n\
\t\t undeclare the predicat as basic event predicat.\n\
\n\
\t undeclare function func\n\
\t\t undeclare the evaluable function func.\n\
\n\
\t unify expr expr \n\
\t\t unify two expressions.\n\
\n\
%s\
\n\
%s", parser_warning, bug_report);
}

void print_intro()
{
     printf(LG_STR("\n\
\t\tBt2f (%s EN).\n\
\n\
\t%s\n\
\n\
Version     : %s\n\
Compiled on : %s\n\
Date        : %s\n\
\n\
 ALL RIGHTS RESERVED\n\
\n\
Bt2f comes with ABSOLUTELY NO WARRANTY.\n\
",
		   "\n\
\t\tBt2f (%s FR).\n\
\n\
\t%s\n\
\n\
Version     : %s\n\
Compiled on : %s\n\
Date        : %s\n\
\n\
 ALL RIGHTS RESERVED\n\
\n\
Bt2f est fourni sans garantie aucune.\n\
"), package_version, COPYRIGHT_STRING, package_version, host, date);
}


static void quit_handler(int sig)
{
     fprintf(stderr,"Quitting on signal %d.\n", sig);
     wrap_up();			/* this is probably not safe...  */
     fprintf(stderr,"Finished cleaning up.\n");
     exit(1);
} 

static struct sigaction quit_act;

void install_quit_handler(void)
{
#ifdef HAVE_SIGACTION
     /* Install the handler. */
     quit_act.sa_handler = quit_handler;
     sigemptyset(&quit_act.sa_mask);
     sigaddset(&quit_act.sa_mask, SIGINT);

     sigaction(SIGINT, &quit_act, NULL);
#else /* SVR4 */
     signal(SIGINT, quit_handler);
#endif /* SVR4 */
}
