%{

/*                               -*- Mode: C -*- 
 * bt2f-parser.l -- Lex parser
 * 
 * Copyright (c) 2024-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
#include <stdlib.h>
#include <sys/file.h>

#include "shashPack_f.h"

#include "constant.h"
#include "macro.h"
#include "opaque.h"
#include "bt2f-type.h"

#include "bt-structure.h"
#include "parser-funct.h"

#include "bt2f-type_f.h"
#include "parser-funct_f.h"

#include "bt2f-parser.h"

#define YY_NO_UNPUT
#define YY_NO_INPUT

/* 
 * This macro allow us to parse from yyin or from a string, and count
 * the lines in the string or in the files.
 */

#define YY_INPUT(buf,result,max_size) \
	if ( YY_CURRENT_BUFFER->yy_is_interactive ) \
		{ \
		int c = '*', n; \
		for ( n = 0; n < max_size && \
			     (c = getc( yyin )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( yyin ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else if ( ((result = fread( buf, 1, max_size, yyin )) == 0) \
		  && (parse_source == PS_FILE) && ferror( yyin ) ) \
		YY_FATAL_ERROR( "input in flex scanner failed" );
%}

ws		[ \t]+
real 		[-\+]?[0-9]+\.[0-9]*
exp		[eE][-\+]?[0-9]+
integer 	[-\+]?[0-9]+[l]?
long_long 	[-\+]?[0-9]+ll
pointer 	0x[0123456789abcdefABCDEF]+
comment		;.*$
qstring 	\"([^"]|\\["])*\" 
untypedid	([^ \t\n~0123456789"&\(\)\$@;\|\.[\]][^ \t\n"\(\)\]]*)
idbar   	(\|[^|]+\|)
nl		\n
or		V
and		&
op		\(
cp		\)
opar		\[
cpar		\]
comma		,
infinity	\.\.\.
oplisp		\(\.
cplisp		\.\)
lvar		\${id}
pvar		@{id}
gpvar		@@{id}

id {untypedid}

     /* Pour montrer qu'on demarre dans cet etat, on l'initialise dans bt2f-parser.y
	Cet etat est necessaire de toutes facons pour accepter des id avec des pname de commandes...*/

%s COMMAND OPF_VERSION NODE_TYPE FIELD_TYPE EDGE_TYPE ACTION_TYPE COLLECT_COMMENT EOFS

%%
{ws}		;
<COMMAND>h	|
<COMMAND>help	|
<COMMAND>\?	{return HELP_TK;}

<COMMAND>echo		{return ECHO_TK;}
<COMMAND>include	{return INCLUDE_TK;}
<COMMAND>load		{return LOAD_TK;}
<COMMAND>[Bb][Tt]	{return BT_TK;}

<COMMAND>set		{return SET_TK;}
<COMMAND>save		{return SAVE_TK;}
<COMMAND>fiacre		{return FCR_TK;}

<COMMAND>condition	{return CONDITION_TK;}

<COMMAND>hippo		{return HIPPO_TK;}
<COMMAND>tina_compact	{return TINA_COMPACT_TK;}
<COMMAND>tick		{return TICK_TK;}

<COMMAND>parallel	{return PARALLEL_TK;}

<COMMAND>on		{return ON_TK;}
<COMMAND>off		{return OFF_TK;}


<COMMAND>quit		|
<COMMAND>q		|
<COMMAND>exit		{return QUIT_TK;}




:name					{return KW_NAME_TK;}
:[Ii][Dd]				{return KW_ID_TK;}
:args					{return KW_ARGS_TK;}
:SFR					{return KW_SFR_TK;}
:SF					{return KW_SF_TK;}
:SR					{return KW_SR_TK;}
:FR					{return KW_FR_TK;}
:S					{return KW_S_TK;}
:F					{return KW_F_TK;}
:R					{return KW_R_TK;}
:num_attempts				{return KW_NUM_ATTEMPTS_TK;}
:num_retries				{return KW_NUM_ATTEMPTS_TK;}
:repeat					{return KW_REPEAT_TK;}
:success				{return KW_SUCCESS_TK;}
:wait					{return KW_WAIT_TK;}
:SV					{return KW_SV_TK;}
:halt					{return KW_HALT_TK;}
:[Ii][Nn][Ii][Tt]					{return KW_INIT_TK;}
:[Mm][Aa][Xx]						{return KW_MAX_TK;}
:[Mm][Ii][Nn]						{return KW_MIN_TK;}
:[Ss][Tt][Aa][Tt][Ee][Ss]				{return KW_STATES_TK;}
:[Tt][Rr][Aa][Nn][Ss][Ii][Tt][Ii][Oo][Nn][Ss]		{return KW_TRANSITIONS_TK;}
:[Aa][Ll][Ll]						{return KW_ALL_TK;}

[Dd][Ee][Ff][Ss][Vv]			{return DEFSV_TK;}
Node					{return NODE_TK;}
BehaviorTree				{return BEHAVIORTREE_TK;}

Action					{return ACTION_TK;}
Condition				{return CONDITION_TK;}

Eval					{return EVAL_TK;}
SetSV					{return SETSV_TK;}

Sequence				{return SEQUENCE_TK;}
AsyncSequence				{return ASYNCSEQUENCE_TK;}
ReactiveSequence			{return REACTIVESEQUENCE_TK;}
SequenceWithMemory			{return SEQUENCEWITHMEMORY_TK;}

Fallback				{return FALLBACK_TK;}
ReactiveFallback			{return REACTIVEFALLBACK_TK;}
AsyncFallback				{return ASYNCFALLBACK_TK;}

Parallel				{return PARALLEL_TK;}
ParallelAll				{return PARALLELALL_TK;}

Repeat					{return REPEAT_TK;}
RetryUntilSuccessful			{return RETRYUNTILSUCCESSFUL_TK;}

IfThenElse				{return IFTHENELSE_TK;}
WhileDoElse				{return WHILEDOELSE_TK;}
Inverter				{return INVERTER_TK;}

RateController				{return RATECONTROLLER_TK;}

Recovery				{return RECOVERY_TK;}
RoundRobin				{return ROUNDROBIN_TK;}
PipelineSequence			{return PIPELINESEQUENCE_TK;}
KeepRunningUntilFailure			{return KEEPRUNNINGUNTILFAILURE_TK;}


ForceSuccess				{return FORCESUCCESS_TK;}
ForceFailure				{return FORCEFAILURE_TK;}

AlwaysSuccess				{return ALWAYSSUCCESS_TK;}
AlwaysFailure				{return ALWAYSFAILURE_TK;}

SubTree					{return SUBTREE_TK;}


<COLLECT_COMMENT>{comment}		{NEWSTR(bt2f_yytext,bt2f_yylval.string);}


^\.\n		{ parser_line_number[parser_index]++;
                  return RESET_DOT_TK;}

{real}		|
{real}{exp}    	{ sscanf(bt2f_yytext,"%lf",&bt2f_yylval.real);
		  return REAL_TK;}
{pointer}       { sscanf(bt2f_yytext+2,"%x", (unsigned int *)&bt2f_yylval.pointer); /* the +2 is to get rid of the 0x... */
		  return POINTER_TK;}
{long_long}       { sscanf(bt2f_yytext,"%lld", &bt2f_yylval.long_long);
		  return LONG_LONG_TK;}
{integer}       { sscanf(bt2f_yytext,"%d", &bt2f_yylval.integer);
		  return INTEGER_TK;}
{comment}	;
{qstring}	{ NEWQSTR(bt2f_yytext,bt2f_yylval.string);
		  return QSTRING_TK;}

\|\|		{ bt2f_yylval.symbol = or_sym;
		  return SYMBOL_TK;}
&		{ bt2f_yylval.symbol = and_sym;
		  return SYMBOL_TK;}
:=		{ bt2f_yylval.symbol = assign_sym;
		  return SYMBOL_TK;}
=		{ bt2f_yylval.symbol = equal_sym;
		  return SYMBOL_TK;}
>=		{ bt2f_yylval.symbol = gte_sym;
		  return SYMBOL_TK;}
\<=		{ bt2f_yylval.symbol = lte_sym;
		  return SYMBOL_TK;}
\<		{ bt2f_yylval.symbol = lt_sym;
		  return SYMBOL_TK;}
>		{ bt2f_yylval.symbol = gt_sym;
		  return SYMBOL_TK;}
~		{ bt2f_yylval.symbol = not_sym;
		  return SYMBOL_TK;}

{lvar}		{ DECLARE_VAR(bt2f_yytext,bt2f_yylval.symbol);
		  return LOGICAL_VAR_TK;}
{pvar} |
{gpvar}		{ DECLARE_VAR(bt2f_yytext,bt2f_yylval.symbol);
		  return PROGRAM_VAR_TK;}

\+>		{  bt2f_yylval.symbol = cond_add_sym;
		  return SYMBOL_TK;}
->		{  bt2f_yylval.symbol = cond_del_sym;
		  return SYMBOL_TK;}
=>		{  bt2f_yylval.symbol = temp_conclude_sym;
		  return SYMBOL_TK;}
~> 		{  bt2f_yylval.symbol = retract_sym;
		  return SYMBOL_TK;}
! 		{  bt2f_yylval.symbol = achieve_sym;
		  return SYMBOL_TK;}
\?	 	{  bt2f_yylval.symbol = test_sym;
		  return SYMBOL_TK;}
#		{  bt2f_yylval.symbol = preserve_sym;
		  return SYMBOL_TK;}
\^	 	{  bt2f_yylval.symbol = twait_sym;
		  return SYMBOL_TK;}
\%		{  bt2f_yylval.symbol = maintain_sym;
		  return SYMBOL_TK;}
\<>		{  bt2f_yylval.symbol = call_sym;
		  return SYMBOL_TK;}


t	|
T		{ bt2f_yylval.symbol = lisp_t_sym;
		  return SYMBOL_TK;}
[nN][iI][lL]	{ bt2f_yylval.symbol = nil_sym;
		  return SYMBOL_TK;}

{id} |
{idbar}		{ DECLARE_ID(bt2f_yytext,bt2f_yylval.symbol);
		  return SYMBOL_TK;}

{op}        	{ return OPENP_TK;}
{cp}        	{ return CLOSEP_TK;}

{nl}	    	{parser_line_number[parser_index]++;}

.		{warning("unexpected character.");}

<EOFS><<EOF>>	 {yyterminate();}
				 
<<EOF>>		 {BEGIN EOFS; return EOF_TK;} /* set the EOF state. */

%%

#include "yy-funct.c"

YY_BUFFER_STATE yy_stdin, yy_ps, yy_mp;
FILE *ps_file, *mp_file;

void take_the_stdin(void)
/* 
 * take_the_stdin - will grab the "real.part" stdin (its value was saved when the process was young 
 *                  and had the descriptor in its memory). We also set i_have_the_stdin so 
 *                  we can prompt correctly...
 */
{
     if (parser_index == MAX_OPEN_INCLUDE) {
	  fprintf(stderr,LG_STR("Too many parser level opened.\n",
				"Trop de niveaux de parsing ouverts.\n"));
	  return;
     }

     i_have_the_stdin = TRUE;

     parser_state_save[parser_index] = YY_CURRENT_BUFFER;
     saved_parse_source[parser_index] = parse_source;

     parser_index++;
     parser_line_number[parser_index] = 1;
     parser_file_name[parser_index] = "stdin"; /* */
     FILE_save[parser_index] = NULL;

     parse_source = PS_FILE;

#if defined(HAVE_SETITIMER) && defined(WANT_TRIGGERED_IO)
     desarm_condition_timer();	/* May generate interrupt which make read on stdin return... */
#endif
     bt2f_yy_switch_to_buffer(yy_stdin);
     printf(LG_STR("Now parsing from stdin, the main loop is suspended.\n",
		   "Parsing du stdin, la boucle principale est suspendue.\n"));
}

void leave_the_stdin(void)
/* 
 * leave_the_stdin - will leave the stdin alone... and will reconnect its stdin to its 
 *                   socket (which value was saved in ps_socket). This is done by
 *                   sending one char down the tube, which instruct the server to grab the 
 *                   stdin again.
 *                   Return void.
 */
{
     if (!i_have_the_stdin) {
	  fprintf(stderr,LG_STR("You do not parse from stdin.\n",
				"Vous ne parsez pas de stdin.\n"));
	  return;
     }
     
     i_have_the_stdin = FALSE;
     parser_file_name[parser_index] = NULL;
     parser_line_number[parser_index] = 0;
     parser_index--;

     yy_switch_to_buffer(parser_state_save[parser_index]);
     parse_source = saved_parse_source[parser_index];

     printf(LG_STR("Quitting stdin parsing, the main loop is resumed.\n",
		   "Arret du parsing du stdin, la boucle principale est reprise.\n"));

#if defined(HAVE_SETITIMER) && defined(WANT_TRIGGERED_IO)
     arm_condition_timer();
#endif
}

void create_yy_mp(void)
{

     /* PROTECT_NULL_SYSCALL_FROM_EINTR(mp_file,fdopen(mp_socket,"r")); */

     /* if (!mp_file) { */
     /* 	  perror("create_yy_mp: fdopen"); */
     /* } */

     /* yy_mp = bt2f_yy_create_buffer(mp_file,YY_BUF_SIZE); */
}

void create_yy_stdin_and_ps(void)
{
     /* yy_stdin = bt2f_yy_create_buffer(stdin,YY_BUF_SIZE); */

     /* PROTECT_NULL_SYSCALL_FROM_EINTR(ps_file,fdopen(ps_socket,"r")); */

     /* if (!ps_file) { */
     /* 	  perror("create_yy_stdin_and_ps: fdopen"); */
     /* } */

     /* yy_ps = bt2f_yy_create_buffer(ps_file,YY_BUF_SIZE); */
}

void yy_switch_to_yy_ps(void)
{
     bt2f_yy_switch_to_buffer(yy_ps);
}

void yy_switch_to_yy_mp(void)
{
     bt2f_yy_switch_to_buffer(yy_mp);
}


void yy_begin_0(void)
{
     BEGIN 0;
}

void yy_begin_COMMAND()
{
     BEGIN COMMAND;
}

void yy_begin_OPF_VERSION(void)
{
     BEGIN OPF_VERSION;
}

void yy_begin_NODE_TYPE(void)
{
     BEGIN NODE_TYPE;
}

void yy_begin_FIELD_TYPE(void)
{
     BEGIN FIELD_TYPE;
}

void yy_begin_ACTION_TYPE(void)
{
     BEGIN ACTION_TYPE;
}

void yy_begin_EDGE_TYPE(void)
{
     BEGIN EDGE_TYPE;
}

void yy_begin_COLLECT_COMMENT(void)
{
     BEGIN COLLECT_COMMENT;
}
