%{
/*                               -*- Mode: C -*- 
 * bt2f-parser.y -- yacc grammaire
 * 
 * Copyright (c) -2023-2024 Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include "config.h"

#include <stdio.h>
#include <sys/file.h>

#include "bt2f-lex.h"

#include "constant.h"
#include "macro.h"
#include "opaque.h"
#include "shashPack.h"
#include "shashPack_f.h"

#include "bt2f-type.h"

#include "bt-structure.h"
#include "bt2f.h"
#include "help_f.h"
#include "parser-funct.h"
#include "bt2f-print.h"
#include "bt2f-sprint.h"

#include "yy-funct_f.h"
#include "parser-funct_f.h"
#include "bt2f-type_f.h"
#include "bt2f-print_f.h"
#include "bt2f-sprint_f.h"
#include "bt-structure_f.h"

char *node_name;
PString type_string;
Slist *text, *graphic, *step_list;
long int read_check_sum;
PBoolean load_ok;


%}

%union {
    double    		real;        /* real value */
    int    		integer;    /* integer value */
    long long int	long_long;    /* long long integer value */
    void    		*pointer;    /* pointer value */
    char    		*string;    /* string buffer */
    Symbol    		symbol;    /* string buffer */
    int    		cmd;        /* command value */
    PBoolean		bool;
    TermList 		list_term;
    ExprList 		list_expr;
    VarList 		var_list;
    Envar       	*var;
    Term 		*term;
    Symbol 		state;
    Transition		*transition;
    Pred_Func_Rec	*pred_func_rec;
    Expression 		*expr;
  //    BTree		*btree;
  //    BTNode_Type		btnode_type;
    BTNode		*btnode;
    BTNodeList 		btnode_list;
    Slist 		*list;
}

%token <real> 		REAL_TK
%token <pointer>	POINTER_TK
%token <integer> 	INTEGER_TK
%token <long_long> 	LONG_LONG_TK
%token <string> 	QSTRING_TK SYMBOL_TK LOGICAL_VAR_TK PROGRAM_VAR_TK

%token <cmd>		EOF_TK SAVE_TK FCR_TK LOAD_TK BT_TK INCLUDE_TK ECHO_TK RESET_DOT_TK HELP_TK
			QUIT_TK SET_TK HIPPO_TK TINA_COMPACT_TK TICK_TK OPENP_TK CLOSEP_TK BEHAVIORTREE_TK
			ON_TK OFF_TK 
			KW_NAME_TK KW_ID_TK KW_SFR_TK KW_SF_TK KW_SR_TK KW_FR_TK KW_S_TK KW_F_TK
			KW_R_TK KW_NUM_ATTEMPTS_TK KW_REPEAT_TK KW_ARGS_TK KW_SUCCESS_TK
			KW_WAIT_TK KW_HALT_TK KW_SV_TK
			ACTION_TK EVAL_TK DEFSV_TK SETSV_TK
			CONDITION_TK SEQUENCE_TK ASYNCSEQUENCE_TK REACTIVESEQUENCE_TK SEQUENCEWITHMEMORY_TK
			FALLBACK_TK REACTIVEFALLBACK_TK ASYNCFALLBACK_TK PARALLEL_TK PARALLELALL_TK
			RETRYUNTILSUCCESSFUL_TK IFTHENELSE_TK WHILEDOELSE_TK INVERTER_TK FORCESUCCESS_TK
			RECOVERY_TK RATECONTROLLER_TK ROUNDROBIN_TK PIPELINESEQUENCE_TK KEEPRUNNINGUNTILFAILURE_TK
			FORCEFAILURE_TK REPEAT_TK ALWAYSSUCCESS_TK ALWAYSFAILURE_TK SUBTREE_TK NODE_TK
			KW_INIT_TK KW_MIN_TK KW_MAX_TK KW_STATES_TK KW_TRANSITIONS_TK KW_ALL_TK
/* %token <cmd>		RESET_TK SEND_TK SET_TK SHOW_TK STAT_TK TRACE_TK UNIFY_TK DB_TK */
/* %token <cmd>		FACT_TK GOAL_TK TSYM_TK INTENTION_TK PSK_TK PSKF_TK KERNEL_TK */
/* %token <cmd>		COMPLETION_TK */
/* %token <cmd>		META_TK EVAL_TK POST_TK PROFILING_TK SILENT_TK FRAME_TK  */
/* %token <cmd>		SOAK_TK THREADING_TK INTEND_TK SUC_FAIL_TK ALL_TK ON_TK OFF_TK  */
/* %token <cmd>		VERSION_TK COPYRIGHT_TK MEMORY_TK FF_TK BE_TK CWP_TK RELEVANT_TK  */
/* %token <cmd>		APPLICABLE_TK TAKE_STDIN_TK DISCONNECT_TK QUIT_TK TEXT_TK  */
/* %token <cmd>		GRAPHIC_TK CALL_EXTERNAL_TK DUMMY_TK SAVE_TK FCR_TK UNDECLARE_TK TYPE_TK */
/* %token <cmd>		EVALUABLE_TK AC_TK EXTERNAL_TK PREDICATE_TK FUNCTION_TK REQUIRE_TK SYM_TK */
/* %token <cmd>		VARIABLE_TK OPRS_DATA_PATH_TK OP_PREDICATE_TK DUMP_TK MAKE_TK WORLD_TK */
/* %token <cmd>		STEP_TK  NEXT_TK HALT_TK RUN_TK STATUS_TK FAILURE_TK USER_TK HIPPO_TK  */
/* %token <cmd>		TINA_COMPACT_TK TICK_TK */
/* %token <cmd>		OPF_VERSION1 OPF_VERSION2 TYPE_SEP_TK */

/* %token <cmd>		ET_GOAL_TK ET_ELSE_TK ET_THEN_TK ET_IF_TK  */

/* %token <cmd>		SPEC_ACT_TK RESET_DOT_TK */
/* %token <cmd>		OR_TK TEMP_CONCLUDE_TK RETRACT_TK */
/* %token <cmd>		COND_ADD_TK COND_DEL_TK */
/* %token <cmd>		OP_LISP_TK CP_LISP_TK OP_TIME_INT_TK CP_TIME_INT_TK OPENP_TK CLOSEP_TK */
/* %token <cmd>		ACHIEVE_TK TEST_TK  PRESERVE_TK MAINTAIN_TK WAIT_TK CALL_TK EOF_TK */

/* %token <cmd>		KW_INVOCATION_TK KW_BODY_TK KW_CALL_TK KW_PRECONDITION_TK KW_INVARIANT_TK KW_TIME_INTERVAL_TK */
/* %token <cmd>		KW_SUCCESS_TK KW_FAILURE_TK KW_INIT_TK KW_MIN_TK KW_MAX_TK KW_STATES_TK KW_TRANSITIONS_TK KW_INTERRUPT_TK */
/* %token <cmd>		KW_POSTCONDITION_TK KW_START_TK */
/* %token <cmd>		KW_DOCUMENTATION_TK KW_ALL_TK KW_GUARD_TK KW_EFFECTS_TK KW_PROGRESS_TK */
/* %token <cmd>		KW_INPUT_TK KW_OUTPUT_TK DEFTYPE_TK DEFDATA_TK KW_MONITOR_TK */

/* %token <cmd>		KW_NAME_TK KW_ID_TK KW_ARGS_TK KW_NUM_ATTEMPTS_TK KW_REPEAT_TK */
/* 			KW_SFR_TK KW_SF_TK KW_SR_TK KW_FR_TK KW_S_TK KW_F_TK KW_R_TK */
/* 			BEHAVIORTREE_TK ACTION_TK CONDITION_TK */
/* 			SEQUENCE_TK ASYNCSEQUENCE_TK REACTIVESEQUENCE_TK SEQUENCEWITHMEMORY_TK */
/* 			FALLBACK_TK REACTIVEFALLBACK_TK ASYNCFALLBACK_TK PARALLEL_TK PARALLELALL_TK */
/* 			RETRYUNTILSUCCESSFUL_TK IFTHENELSE_TK WHILEDOELSE_TK INVERTER_TK FORCESUCCESS_TK */
/* 			FORCEFAILURE_TK REPEAT_TK ALWAYSSUCCESS_TK ALWAYSFAILURE_TK SUBTREE_TK  */

/* %token <cmd>		WHILE_TK DO_TK PAR_TK IF_TK ELSE_TK GOTO_TK BREAK_TK ELSEIF_TK LABEL_TK  */

%type <cmd>		reset_or_eof
%type <btnode>		BehaviorTree
%type <btnode> 		bt_node bt_node_interm
%type <list>		bt_list new_bt_list top_list key_val_plist key_val_list sv_states states_list sv_transitions
%type <bool> 		on_off_flag
%type <string> 		/* documentation */ read_file_name write_file_name
%type <symbol> 		symbol
%type <pred_func_rec>	pred_func
%type <list_term> 	term_list
%type <var> 		variable
%type <term> 		term
%type <state>		state
%type <transition>	transition
%type <expr> 		expr /* top_expr */
/* %type <list_expr> 	list_par_expr list_expr */
/* %type <var_list> 	var_list */
%%
one_command_return: 				
	command 			{YYACCEPT;} /* We have to return explicitly to free the memory
						       and get out of here*/
	| EOF_TK			{YYACCEPT;}
	;

command:
        SAVE_TK FCR_TK				{yy_begin_0();} 
		write_file_name			{write_fcr($4); FREE($4);}
	| LOAD_TK BT_TK				{yy_begin_0();}
		read_file_name 		        {if (!be_quiet) printf("%sLoading BTF File \"%s\".\n",parser_indent, $4);
						 if (open_file_for_yyparse($4)<0) {
						      printf("%sError, Abort  loading BTF File.\n",parser_indent);
						      FREE($4);
						      YYABORT;
						 }
						 btn_stack = sl_make_slist(); /* current_graph_file_name = $4; */} 
		bt_list 			{close_file_for_yyparse(); FREE($4);
						 if (!be_quiet) printf("%sLoading BTF File done.\n",parser_indent);}
	| INCLUDE_TK   				{yy_begin_0();} 
		read_file_name 			{if (open_file_for_yyparse($3)<0) YYABORT; FREE($3);} 
	| ECHO_TK 				{enable_variable_parsing();}
		echo_tout			{disable_variable_parsing();}
	| ECHO_TK error RESET_DOT_TK		{warning("Parsing error, expecting a expr, expr or expr after \"echo\"");}
	| HELP_TK				{print_help_from_parser();}
	| QUIT_TK				{quit = TRUE;}
	| SET_TK HIPPO_TK on_off_flag		{bt2f_option[HIPPO]=$3;}
	| SET_TK TINA_COMPACT_TK on_off_flag		{bt2f_option[TINA_COMPACT]=$3;}
	| SET_TK TICK_TK INTEGER_TK		{ns_per_tick = $3;}
	| SET_TK TICK_TK ACTION_TK		{htick = ACTION;}
	| SET_TK TICK_TK BT_TK			{htick = BT;}
	| SET_TK TICK_TK NODE_TK		{htick = NODE;}
	| RESET_DOT_TK				{}
	| error reset_or_eof			{warning("Parsing error, unknown command, resetting the parser");}
        ;


symbol: SYMBOL_TK {$$=$1;}
//	| SYMBOL_TK TYPE_SEP_TK SYMBOL_TK 	{set_type_by_name($3,$1);$$=$1;}
	;

/* add_frame: */
/* 	ADD_TK 					{enable_variable_parsing(); yy_begin_0();} /
/* 	; */

echo_tout:					{check_pfr = FALSE;}
        expr					{print_expr($2); free_expr($2); printf("\n");
						 check_pfr = TRUE;}
	;


reset_or_eof:  RESET_DOT_TK
	| EOF_TK
	;

bt_list:					{yy_begin_OPF_VERSION();}
	new_bt_list				{$$=$2;}
	| error	EOF_TK				{warning("Either this is not a BTF file, or this BTF has some errors (see above)."); exit(1);}

	;

new_bt_list: /* empty  file... */		{$$=sl_make_slist();}
	| OPENP_TK 				{yy_begin_0(); check_pfr = FALSE; /* make_btset(); */} 
	  top_list 
	  CLOSEP_TK				{$$=$3;}
	;

top_list: 
/* empty */				{bt_list=sl_make_slist();
	    				 btnode_list=sl_make_slist(); 
					 sv_list = sl_make_slist();
					 btidactarg_list = sl_make_slist();}
	| top_list BehaviorTree			{sl_add_to_tail(btnode_list, $2); sl_add_to_tail(bt_list, $2); $$=bt_list;}
	| top_list sv
	;


sv: OPENP_TK DEFSV_TK SYMBOL_TK 		{make_sv($3);} 
		sv_fields_list 		{add_sv_to_sv_list();}
		 CLOSEP_TK
;

sv_fields_list: sv_field 
	   | sv_fields_list sv_field
;

sv_field: sv_states_field
     | sv_init_field
     | sv_min_field
     | sv_max_field
     | sv_transitions_field
;


sv_states_field: KW_STATES_TK sv_states
		{
		  current_sv->states = $2;
		}
;

sv_init_field: KW_INIT_TK sv_init
;

sv_min_field: KW_MIN_TK INTEGER_TK
		{
		  current_sv->min = $2;
		}
;

sv_max_field: KW_MAX_TK INTEGER_TK
		{
		  current_sv->max = $2;
		}
;



sv_transitions_field: KW_TRANSITIONS_TK OPENP_TK sv_transitions CLOSEP_TK
		{
		  current_sv->transitions = $3;
		}
		| KW_TRANSITIONS_TK KW_ALL_TK
		{
		  current_sv->all_transitions = TRUE;
		};


sv_states: OPENP_TK states_list CLOSEP_TK	{$$ = $2;}
	;


sv_init: state
		{current_sv->init = $1;}
	| INTEGER_TK
		{current_sv->initi = $1;}
	;

state: SYMBOL_TK		{$$ = $1;}
	;

states_list:
        state 					{$$=sl_make_slist(); sl_add_to_tail($$,$1);}
	| states_list state 			{sl_add_to_tail($1,$2);$$=$1;}
	;

sv_transitions: transition			{$$=sl_make_slist(); sl_add_to_tail($$,$1);}
        | sv_transitions transition		{sl_add_to_tail($1,$2);$$=$1;}
	;

transition: OPENP_TK state state CLOSEP_TK	{$$ = make_transition($2, $3);}

BehaviorTree:  OPENP_TK BEHAVIORTREE_TK 	{sl_add_to_head(btn_stack,make_btnode_type(BehaviorTree)); } 
		bt_fields_list CLOSEP_TK 	{$$=current_btnode(); sl_get_from_head(btn_stack);}
;

on_off_flag:
	 ON_TK 					{$$ = TRUE;}
	|OFF_TK					{$$ = FALSE;}
	   ;

name_field: KW_NAME_TK symbol			{current_btnode()->name = $2;}
;

sv_field: KW_SV_TK symbol			{current_btnode()->sv = $2;}
;

eval_expr:  expr				{current_btnode()->expr = $1;}
;

id_field:  KW_ID_TK symbol			{current_btnode()->id = $2;}
;

rt_field:  KW_SFR_TK 			{current_btnode()->PossibleRet = SFR;}
	| KW_SF_TK			{current_btnode()->PossibleRet = SF;}
	| KW_SR_TK			{current_btnode()->PossibleRet = SR;}
	| KW_FR_TK			{current_btnode()->PossibleRet = FR;}
	| KW_S_TK			{current_btnode()->PossibleRet = S;}
	| KW_F_TK			{current_btnode()->PossibleRet = F;}
	| KW_R_TK			{current_btnode()->PossibleRet = R;}
;

num_attempts_field:  KW_NUM_ATTEMPTS_TK INTEGER_TK	{current_btnode()->u.num_attempts = $2;}
;

repeatkw_field:  KW_REPEAT_TK INTEGER_TK	{current_btnode()->u.repeat = $2;}
;

args_field:  KW_ARGS_TK 		{enable_variable_parsing();}
		key_val_plist		{disable_variable_parsing(); 
					 current_btnode()->args = $3;}
;

needed_success_field:  KW_SUCCESS_TK INTEGER_TK			{current_btnode()->u.needed_success = $2;}
;

halt_field:  KW_HALT_TK INTEGER_TK			{current_btnode()->halt = $2;}
;

wait_field:  KW_WAIT_TK INTEGER_TK			{current_btnode()->wait = $2;}
;

key_val_plist:  OPENP_TK key_val_list CLOSEP_TK 	{$$ = $2;}

key_val_list:
	/* empty */				{$$=sl_make_slist();}
	| key_val_list symbol term		{sl_add_to_tail($1,make_key_val($2,$3)); $$=$1;}

bt_fields_list: bt_field 
	| bt_fields_list bt_field
;

bt_field: name_field
	| rt_field
     	| id_field
     	| args_field
	| bt_node
;

bt_node: bt_node_interm 			{sl_add_to_tail(current_btnode()->children,$1);
	   					 sl_add_to_tail(btnode_list, $1);}
;

bt_node_interm: 
	OPENP_TK ACTION_TK 			{sl_add_to_head(btn_stack,make_btnode_type(Action)); } 
		action_fields_list CLOSEP_TK 	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK CONDITION_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Condition)); } 
 		condition_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK EVAL_TK 			{sl_add_to_head(btn_stack,make_btnode_type(Eval)); } 
 		eval_field CLOSEP_TK		{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK SETSV_TK 			{sl_add_to_head(btn_stack,make_btnode_type(SetSV)); } 
 		setsv_fields_list CLOSEP_TK		{$$=current_btnode(); sl_get_from_head(btn_stack);}
      	| OPENP_TK SEQUENCE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Sequence)); } 
		 sequence_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK ASYNCSEQUENCE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(AsyncSequence)); } 
		 asyncsequence_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK REACTIVESEQUENCE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(ReactiveSequence)); } 
		 reactivesequence_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK SEQUENCEWITHMEMORY_TK 	{sl_add_to_head(btn_stack,make_btnode_type(SequenceWithMemory)); } 
		 sequencewithmemory_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK FALLBACK_TK 			{sl_add_to_head(btn_stack,make_btnode_type(Fallback)); } 
		 fallback_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK REACTIVEFALLBACK_TK 		{sl_add_to_head(btn_stack,make_btnode_type(ReactiveFallback)); } 
		 reactivefallback_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK ASYNCFALLBACK_TK 		{sl_add_to_head(btn_stack,make_btnode_type(AsyncFallback)); } 
		 asyncfallback_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK PARALLEL_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Parallel)); } 
		 parallel_fields_list CLOSEP_TK	{$$=current_btnode(); set_needed_success_if_not_set($$);
		   				 sl_get_from_head(btn_stack);}
	| OPENP_TK PARALLELALL_TK 		{sl_add_to_head(btn_stack,make_btnode_type(ParallelAll)); } 
		 parallelall_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK RETRYUNTILSUCCESSFUL_TK 		{sl_add_to_head(btn_stack,make_btnode_type(RetryUntilSuccessful)); } 
		 retryuntilsuccessful_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK IFTHENELSE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(IfThenElse)); } 
		 ifthenelse_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK WHILEDOELSE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(WhileDoElse)); } 
		 whiledoelse_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK INVERTER_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Inverter)); } 
		 inverter_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK RECOVERY_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Recovery)); } 
		 recovery_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK RATECONTROLLER_TK 		{sl_add_to_head(btn_stack,make_btnode_type(RateController)); } 
		 ratecontroller_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK ROUNDROBIN_TK 		{sl_add_to_head(btn_stack,make_btnode_type(RoundRobin)); } 
		 roundrobin_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK PIPELINESEQUENCE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(PipelineSequence)); } 
		 pipelinesequence_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK KEEPRUNNINGUNTILFAILURE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(KeepRunningUntilFailure)); } 
		 keeprunninguntilfailure_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK FORCESUCCESS_TK 		{sl_add_to_head(btn_stack,make_btnode_type(ForceSuccess)); } 
		 forcesuccess_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK FORCEFAILURE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(ForceFailure)); } 
		 forcefailure_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK REPEAT_TK 		{sl_add_to_head(btn_stack,make_btnode_type(Repeat)); } 
		 repeat_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK ALWAYSSUCCESS_TK 		{sl_add_to_head(btn_stack,make_btnode_type(AlwaysSuccess)); } 
		 alwayssuccess_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK ALWAYSFAILURE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(AlwaysFailure)); } 
		 alwaysfailure_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	| OPENP_TK SUBTREE_TK 		{sl_add_to_head(btn_stack,make_btnode_type(SubTree)); } 
		 subtree_fields_list CLOSEP_TK	{$$=current_btnode(); sl_get_from_head(btn_stack);}
	;

setsv_fields_list: setsv_fields
	| setsv_fields_list setsv_fields
;

setsv_fields: name_field
	| sv_field
	| id_field
     	| args_field
;

action_fields_list: action_fields
	| action_fields_list action_fields
;

action_fields: name_field
	| rt_field
	| id_field
     	| args_field
;

condition_fields_list: condition_fields
	| condition_fields_list condition_fields
;

condition_fields: name_field
     	| args_field
	| rt_field
	| id_field
;

eval_field:  			{enable_variable_parsing();}
	 	eval_expr 	{disable_variable_parsing();}
;

sequence_fields_list: sequence_fields
	| sequence_fields_list sequence_fields
;

sequence_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

asyncsequence_fields_list: asyncsequence_fields
	| asyncsequence_fields_list asyncsequence_fields
;

asyncsequence_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

reactivesequence_fields_list: reactivesequence_fields
	| reactivesequence_fields_list reactivesequence_fields
;

reactivesequence_fields: name_field
     	| args_field
	| rt_field
	| halt_field
	| id_field
	| bt_node
;

sequencewithmemory_fields_list: sequencewithmemory_fields
	| sequencewithmemory_fields_list sequencewithmemory_fields
;

sequencewithmemory_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

fallback_fields_list: fallback_fields
	| fallback_fields_list fallback_fields
;

fallback_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

reactivefallback_fields_list: reactivefallback_fields
	| reactivefallback_fields_list reactivefallback_fields
;

reactivefallback_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

asyncfallback_fields_list: asyncfallback_fields
	| asyncfallback_fields_list asyncfallback_fields
;

asyncfallback_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

parallel_fields_list: parallel_fields
	| parallel_fields_list parallel_fields
;

parallel_fields: name_field
     	| args_field
     	| needed_success_field
     	| halt_field
     	| wait_field
	| rt_field
	| id_field
	| bt_node
;

parallelall_fields_list: parallelall_fields
	| parallelall_fields_list parallelall_fields
;

parallelall_fields: name_field
     	| args_field
     	| halt_field
     	| wait_field
	| rt_field
	| id_field
	| bt_node
;

retryuntilsuccessful_fields_list: retryuntilsuccessful_fields
	| retryuntilsuccessful_fields_list retryuntilsuccessful_fields
;

retryuntilsuccessful_fields: name_field
     	| num_attempts_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

ifthenelse_fields_list: ifthenelse_fields
	| ifthenelse_fields_list ifthenelse_fields
;

ifthenelse_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

whiledoelse_fields_list: whiledoelse_fields
	| whiledoelse_fields_list whiledoelse_fields
;

whiledoelse_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

inverter_fields_list: inverter_fields
	| inverter_fields_list inverter_fields
;

inverter_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

recovery_fields_list: recovery_fields
	| recovery_fields_list recovery_fields
;

recovery_fields: name_field
     	| args_field
     	| num_attempts_field
	| rt_field
	| id_field
	| bt_node
;

ratecontroller_fields_list: ratecontroller_fields
	| ratecontroller_fields_list ratecontroller_fields
;

ratecontroller_fields: name_field
     	| args_field
     	| num_attempts_field
	| rt_field
	| id_field
	| bt_node
;

roundrobin_fields_list: roundrobin_fields
	| roundrobin_fields_list roundrobin_fields
;

roundrobin_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

keeprunninguntilfailure_fields_list: keeprunninguntilfailure_fields
	| keeprunninguntilfailure_fields_list keeprunninguntilfailure_fields
;

keeprunninguntilfailure_fields: name_field
     	| args_field
	| id_field
	| bt_node
;

pipelinesequence_fields_list: pipelinesequence_fields
	| pipelinesequence_fields_list pipelinesequence_fields
;

pipelinesequence_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

forcesuccess_fields_list: forcesuccess_fields
	| forcesuccess_fields_list forcesuccess_fields
;

forcesuccess_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

forcefailure_fields_list: forcefailure_fields
	| forcefailure_fields_list forcefailure_fields
;

forcefailure_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

repeat_fields_list: repeat_fields
	| repeat_fields_list repeat_fields
;

repeat_fields: name_field
     	| repeatkw_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;

alwayssuccess_fields_list: alwayssuccess_fields
	| alwayssuccess_fields_list alwayssuccess_fields
;

alwayssuccess_fields: name_field
     	| args_field
	| rt_field
	| id_field
;

alwaysfailure_fields_list: alwaysfailure_fields
	| alwaysfailure_fields_list alwaysfailure_fields
;

alwaysfailure_fields: name_field
     	| args_field
	| rt_field
	| id_field
;

subtree_fields_list: subtree_fields
	| subtree_fields_list subtree_fields
;

subtree_fields: name_field
     	| args_field
	| rt_field
	| id_field
	| bt_node
;


/* top_expr: expr				{$$=build_inst($1);} */
/* ; */


/* documentation:			{$$ = NULL;} */
/* 	| QSTRING_TK		{$$ = $1;} */
/* 	; */

variable:
	LOGICAL_VAR_TK 				{$$=build_var_in_frame($1,LOGICAL_VARIABLE);}
	/* | LOGICAL_VAR_TK  TYPE_SEP_TK SYMBOL_TK	{$$=build_var_in_frame($1,LOGICAL_VARIABLE); */
	/* 					 set_var_type_by_name($3, $$);} */
	| PROGRAM_VAR_TK			{$$=build_var_in_frame($1,PROGRAM_VARIABLE);}
	/* | PROGRAM_VAR_TK  TYPE_SEP_TK SYMBOL_TK	{$$=build_var_in_frame($1,PROGRAM_VARIABLE); */
	/* 					 set_var_type_by_name($3, $$);} */
	;

/* list_par_expr: */
/* 	OPENP_TK list_expr CLOSEP_TK 		{$$=$2;} */
/* 	| OPENP_TK CLOSEP_TK 			{$$=sl_make_slist();} */
/* 	; */

/* list_expr: */
/* 	expr 					{$$=sl_make_slist(); sl_add_to_tail($$,$1);} */
/* 	| list_expr expr 			{sl_add_to_tail($1,$2);$$=$1;} */
/* 	; */

expr:
        OPENP_TK pred_func term_list CLOSEP_TK 	{$$=build_expr_pfr_terms($2,$3); 
	                                         if (! $$) {warning("badly formed expression"); YYABORT;}}
	/* |  OPENP_TK pred_func error CLOSEP_TK		{warning("badly formed expression"); YYABORT;} */
	;

pred_func: SYMBOL_TK				{$$ = find_or_create_pred_func($1);}
//	| variable				{$$ = create_var_pred_func($1);}
	;
        
term_list:
	/*nothing*/ 				{$$=(TermList)sl_make_slist();}
	| term_list term 			{$$=build_term_list($1,$2);}
	;


term:
	symbol 				{$$=build_id($1);}
        | INTEGER_TK 				{$$=build_integer($1);}
	/* | LONG_LONG_TK  			{$$=build_long_long($1);} */
	/* | POINTER_TK  				{$$=build_pointer($1);} */
	| REAL_TK  				{$$=build_float($1);}
	| QSTRING_TK				{$$=build_qstring($1);}
	| variable  				{$$=build_term_from_var($1);}
        | expr	 				{$$=build_term_expr($1);}
	;

/* var_list: */
/* 	variable 				{$$=(VarList)sl_make_slist(); sl_add_to_head($$,$1);} */
/* 	| var_list variable			{sl_add_to_tail($1,$2); $$=$1;} */
/* 	; */

/* var_type_list: */
/* 	variable SYMBOL_TK			//{$$=(VarList)sl_make_slist(); sl_add_to_head($$,$1);} */
/* 	| var_type_list variable SYMBOL_TK	//	{sl_add_to_tail($1,$2); $$=$1;} */
/* 	; */


read_file_name: QSTRING_TK			{$$=canonicalize_read_file_name($1);}
	;

write_file_name: QSTRING_TK			{$$=$1;}
	;

%%

PBoolean loaded_include_file(char *filename)
{
	 return (current_bt2f->included_file_name ?
	     (sl_search_slist(current_bt2f->included_file_name, filename, equal_string) != NULL) :
	     FALSE);
}

void provide_include_file(char *filename)
{
  if (! current_bt2f->included_file_name)  
    current_bt2f->included_file_name = sl_make_slist();
  
  sl_add_to_tail(current_bt2f->included_file_name,filename);
}
