
/*                               -*- Mode: C -*- 
 * bt2f.c -- The big boss
 * 
 * Copyright (c) 1991-2024 Francois Felix Ingrand.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/socket.h>

#include <netinet/in.h>

#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

#include <unistd.h>
#include <netdb.h>


#include "bt2f.h"
#include "bt2f-main.h"
#include "bt2f-pred-func.h"

#include "top-lev.h"


#include "bt2f_f.h"
#include "bt2f-type_f.h"
#include "bt2f-print_f.h"
#include "bt2f-sprint.h"
#include "bt2f-sprint_f.h"
#include "bt2f-pred-func_f.h"
#include "top-lev_f.h"

Bt2f *current_bt2f;

int ps_socket;

void declare_predicat()
{
     declare_pred_from_symbol(soak_sym);	/* Declare the soak symbol */
     declare_pred_from_symbol(app_ops_fact_sym);
     declare_pred_from_symbol(fact_invoked_ops_sym);
     declare_pred_from_symbol(db_satisf_goal_sym);
     declare_pred_from_symbol(app_ops_goal_sym);
     declare_pred_from_symbol(goal_invoked_ops_sym);
     declare_pred_from_symbol(failed_goal_sym);
     declare_pred_from_symbol(failed_sym);
     declare_pred_from_symbol(request_sym);
     declare_pred_from_symbol(achieved_sym);
     declare_pred_from_symbol(wake_up_intention_sym);
     declare_pred_from_symbol(read_response_id_sym);
     declare_pred_from_symbol(read_response_sym);
     declare_pred_from_symbol(assign_sym);
     declare_pred_from_symbol(equal_sym);
     declare_pred_from_symbol(gte_sym);
     declare_pred_from_symbol(lte_sym);
     declare_pred_from_symbol(gt_sym);
     declare_pred_from_symbol(lt_sym);
     declare_pred_from_symbol(or_sym);
     declare_pred_from_symbol(and_sym);
     declare_pred_from_symbol(not_sym);
     declare_pred_from_symbol(cond_add_sym);
     declare_pred_from_symbol(cond_del_sym);
     declare_pred_from_symbol(temp_conclude_sym);
     declare_pred_from_symbol(retract_sym);
     declare_pred_from_symbol(achieve_sym);
     declare_pred_from_symbol(test_sym);
     declare_pred_from_symbol(preserve_sym);
     declare_pred_from_symbol(twait_sym);
     declare_pred_from_symbol(maintain_sym);
     declare_pred_from_symbol(call_sym);
}

void declare_be_pred()
{
}

void free_bt2f(Bt2f *bt2f)
{

     FREE_SLIST(bt2f->new_facts);

     FREE_SLIST(bt2f->locked_messages);
     FREE_SLIST(bt2f->facts);

     FREE_SLIST(bt2f->new_goals);
     FREE_SLIST(bt2f->goals);

     FREE_SLIST(bt2f->concluded_be);
     if (bt2f->included_file_name) 
         FREE_SLIST(bt2f->included_file_name);
     
     FREE_OBJECT(bt2f);
}

char *kernel_name()
{
     return current_bt2f->name;
}

Bt2f *make_bt2f(PString name)
{
/* 
 * make_bt2f - Create and return a bt2f object named "name".
 *            It creates all the necessary objects present in a BT2F agent.
 */
     Bt2f *bt2f=MAKE_OBJECT(Bt2f);

     bt2f->name = name;
     bt2f->types = sl_make_slist();

     /* bt2f->database =  make_database(); */

     bt2f->posted_meta_fact = FALSE;
     bt2f->critical_section = FALSE;
     //     bt2f->intention_graph = make_intention_graph();
     //     bt2f->relevant_op =  make_relevant_op(); 
     bt2f->new_facts = sl_make_slist();
     bt2f->locked_messages = sl_make_slist();
     bt2f->facts = sl_make_slist();
     bt2f->new_goals = sl_make_slist();
     bt2f->goals = sl_make_slist();
     bt2f->conditions_list = sl_make_slist();
     bt2f->concluded_be = sl_make_slist();
     bt2f->included_file_name = NULL;
     return bt2f;
}

TermList make_external_term_list(int nb_arg, ...)
{
     va_list list_arg;
     int param_cour;
     TermList param_list = sl_make_slist();

     va_start(list_arg, nb_arg);
  
     for(param_cour = 1; param_cour <= nb_arg; param_cour++) {
	  Term_Type type;
	  type = va_arg(list_arg, Term_Type);

	  switch(type) {
	  case TT_FLOAT:
	       add_to_tail(param_list, build_float(va_arg(list_arg, double)));
	       break;
	  case INTEGER:
	       add_to_tail(param_list, build_integer(va_arg(list_arg, int)));
	       break;
	  case LONG_LONG:
	       add_to_tail(param_list, build_long_long(va_arg(list_arg, long long int)));
	       break;
	  case STRING:
	       add_to_tail(param_list, build_string(va_arg(list_arg, char *)));
	       break;
	  case EXPRESSION:
	       add_to_tail(param_list, build_term_expr(va_arg(list_arg, Expression *)));
	       break;
	  case U_POINTER:
	       add_to_tail(param_list, build_pointer(va_arg(list_arg, void *)));
	       break;
	  case TT_ATOM:
	       add_to_tail(param_list, 
			   build_id(find_atom(va_arg(list_arg, char *))));
	       break;
	  /* case LISP_LIST: */
	  /*      add_to_tail(param_list, build_l_list(va_arg(list_arg,  L_List))); */
	  /*      break; */
	  case U_MEMORY:
	  default:
	       fprintf(stderr,"make_external_term_list: Unknown/Unsupported type.\n");
	       va_end(list_arg);
	       return param_list;
	  }
     }
  
     va_end(list_arg);
     return param_list;
}

Expression *make_external_expr(char *function, TermList param_list)
{
     Pred_Func_Rec *func = find_or_create_pred_func(declare_atom(function));

     return build_expr_pfr_terms(func,param_list);
}
