/*
 * Copyright (c) 2023-2024 Felix Ingrand, CNRS/LAAS
 *
 * All rights reserved.
 * 
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source  code must retain the  above copyright
 *    notice and this list of conditions.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice and  this list of  conditions in the  documentation and/or
 *    other materials provided with the distribution.
 * 
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <getopt.h>
#include <inttypes.h>


/* #include "portLib.h" */
/* #include "h2initGlob.h" */
/* #include "taskLib.h" */

#include "hippo.h"

#include "bt2f-hippo.h"
#include "hippo_app_functions.h"

int graph_fd = -1;

static int sec_start = 0;
unsigned int fm_trace_level = 5;
unsigned int sleep_before_start = 0;
extern unsigned long long int global_logical_time;

extern long int (*thread_starting_hook) (const char *thread_name);

void initial(void);

int bt2f_hippo_debug = 0;

extern int hippo_debug;

int graphic_trace = 0;

void h_log_info(const char *format, ...)
{
  va_list va;

  if (! bt2f_hippo_debug) return;

  struct timeval now;

  gettimeofday(&now, NULL);
#ifdef __APPLE__
  printf("%4ld.%06d\t\t[*** Hippo ***] ", now.tv_sec - sec_start, now.tv_usec);
#else
  printf("%4ld.%06ld\t\t[*** Hippo ***] ", now.tv_sec - sec_start, now.tv_usec);
#endif
  va_start(va, format);
  vprintf(format, va);
  va_end(va);
  printf("\n");
}


void uc_log_info(const char *format, ...) /* user code */
{
  va_list va;
  struct timeval now;

  gettimeofday(&now, NULL);
#ifdef __APPLE__
  printf("%4ld.%06d\t\t[*** UserCode ***] ", now.tv_sec - sec_start, now.tv_usec);
#else
  printf("%4ld.%06ld\t\t[*** UserCode ***] ", now.tv_sec - sec_start, now.tv_usec);
#endif
  va_start(va, format);
  vprintf(format, va);
  va_end(va);
  printf("\n");
}


void fm_log_info(const char *format, ...)
{
  va_list va;
  struct timeval now;

  gettimeofday(&now, NULL);
#ifdef __APPLE__
  printf("%4ld.%06d %6lld [*** FiacreModel ***] ", now.tv_sec - sec_start, now.tv_usec, global_logical_time);
#else
    printf("%4ld.%06ld %6lld [*** FiacreModel ***] ", now.tv_sec - sec_start, now.tv_usec, global_logical_time);
#endif
  va_start(va, format);
  vprintf(format, va);
  va_end(va);
  printf("\n");
}

void user_print(const char *format, ...)
{
  printf("[*** User Print ***] ");
    va_list va;
  va_start(va, format);
  vprintf(format, va);
  va_end(va);
  printf("\n");
}

int print_trace_c(int index)
{
  if (fiacre_trace_string[index].level <= fm_trace_level)
    fm_log_info("%s", fiacre_trace_string[index].string);
  return 0;
}

int print_user_print_c(int index)
{
  user_print("%s", fiacre_user_print_string[index].string);
  return 0;
}



void update_graphic_tick(void)
{
  if (graph_fd > 0) {		/* this is a terrible hack... we know this is the second
				   line of the first node... where we print the tick value */
    dprintf(graph_fd, "$c itemconfigure 3 -text %04lld\n", global_logical_time);
  }
}

int graphic_trace_c(int tcldot_index, node_status_c status)
{
  static unsigned long long cached_tick = 0;

  if (cached_tick !=  global_logical_time) {
    cached_tick = global_logical_time;
    update_graphic_tick();
  }

  char *color = "white";
  
  if (graph_fd > 0) {
    switch (status) {
    case ns_failure:
      color = "red";
      break;
    case ns_halt:
      color = "pink";
      break;
    case ns_halted:
      color = "magenta";
      break;
    case ns_idle:
      break;
    case ns_running:
      color = "yellow";
      break;
      break;
    case ns_success:
      color = "chartreuse";
      break;
    case ns_tick_passed:
      color = "turquoise1";
      break;
    case ns_ticked:
      color = "{cornflower blue}";
      break;
    default:
      break;
    }
    
    dprintf(graph_fd, "color_btn %d %s\n", tcldot_index, color); 
  }    
  
  return 0;
}

static const char shortopts_string[] = "PF:s:G:Hh";
static struct option longopts_list[] = {
  { "BT2F_debug",	optional_argument,	NULL,	'P'},
  { "Fiacre_trace",	required_argument,	NULL,	'F'},
  { "sleep",		required_argument,	NULL,	's'},
  { "Graphic_trace",	required_argument,	NULL,	'G'},
  { "hippo_debug",	no_argument,		NULL,	'H'},
  { "help",		no_argument,		NULL,	'h'},
  { NULL, 0, NULL, 0}
};

/* options */
static const char usage_string[] = "\
  -P,--Bt2f_debug\tBt2f/Hippo debuging\n\
  -F,--Fiacre_trace\tFiacre model execution trace level\n\
  -s,--sleep\tSleep n seconds before starting the engine\n\
  -G,--Graphic_trace\tBT graphic_trace\n\
  -H,--hippo_trace\tHippo engine execution trace (very verbose)\n\
  -h,--help\t\tprint usage summary (this text)\n\
";

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel,
	  "Usage: %s\n%s\n",  argv0, usage_string);
}

int
main(int argc, char **argv)
{
  char *e;
  int c;
  struct timeval now;

  gettimeofday(&now, NULL);

  sec_start = now.tv_sec;

  thread_starting_hook = NULL;

  char *dot_filename;
  
  /* parse command line options */
  while ((c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1) {
    switch (c) {

    case 'F': {
      unsigned long s = strtoul(optarg, &e, 0);
      if (*e != 0) {
	printf("bad fiacre verbose level: %s\n", optarg);
	exit(1);
      }
      fm_trace_level = s;
      fm_log_info("FiacreModel verbose level: %u", fm_trace_level);
      break;
    }

    case 's': {
      unsigned long sl = strtoul(optarg, &e, 0);
      if (*e != 0) {
	printf("bad number of second: %s\n", optarg);
	exit(1);
      }
      sleep_before_start = sl;
      fm_log_info("FiacreModel sleeping: %u",  sleep_before_start);
      break;
    }

    case 'P': {
      bt2f_hippo_debug = 1;
      break;
    }

    case 'H': {
      hippo_debug++;
      break;
    }
    case 'G': {
      dot_filename =  malloc(strlen(optarg) + 1);

      if (! dot_filename) {
	fprintf(stderr, "malloc: main: dot_filename failed...");
	exit (1);
      }	  
      memcpy(dot_filename,optarg,strlen(optarg) + 1);
      graphic_trace = 1;
      break;
    }

    case 'h':  usage(stdout, argv[0]);  exit(0); break;

    case '?':
    default:
      usage(stderr, argv[0]);
      exit(1);
      break;
    }
  }
  argc -= optind;
  argv += optind;

  if (graphic_trace) {
    int fds[2];                      // an array that will hold two file descriptors
    
    if (pipe(fds) == -1) {
      fprintf(stderr, "Pipe Failed");
      return 1;
    }
    
  
    pid_t pid = fork();
    
    if (pid == -1) {
      fprintf(stderr, "fork Failed");
      return 1;
    }
  
    if (pid == 0) {                  // if pid == 0, then this is the child process
      dup2(fds[0], STDIN_FILENO);    // fds[0] (the read end of pipe) donates its data to file descriptor 0
      close(fds[0]);                 // file descriptor no longer needed in child since stdin is a copy
      close(fds[1]);                 // file descriptor unused in child
      char *argv[] = {(char *)"tclsh", NULL};   // create argument vector
      if (execvp(argv[0], argv) < 0) exit(0);  // run command (exit if something went wrong)
    } 

    // if we reach here, we are in parent process
    close(fds[0]);                 // file descriptor unused in parent
    graph_fd = fds[1]; 

/*     fprintf(stderr, "set dot_filename %s\n\ */
/* source testTcldot.tcl\n\ */
/* source tesdraw.tcl\n\ */
/* map_node_tag\n", dot_filename);  */

    dprintf(graph_fd, "set dot_filename %s\n\
source testTcldot.tcl\n\
source tesdraw.tcl\n\
map_node_tag\n", dot_filename); 

    // send EOF so child can continue (child blocks until all input has been processed):
    //  close(fds[1]); 

  }

  srand(time(NULL));
  
  //  sleep(1); 			/* let the localization thread start */
    
  initial(); 			/* defined in the fiacre synthesized code. */
  sleep(sleep_before_start);
  engine_run();
  
  printf("Shutdown Drone Bt2f/Hippo.\n");
  
  return (EXIT_SUCCESS);
}
