#include <iostream>
#include <cstdio>
#include <string>
#include <sstream>
#include <regex>
#include <vector>
#include <functional>

using namespace std;

bool binary, digraph, xychange;

struct line {int off; string str;};

struct RTnode {
  unsigned label;
  string * content, * attnode, * attedge;
  vector<RTnode *> * desc;
};

struct BTnode {
  unsigned label;
  string * content, * attnode, * attedge;
  BTnode * f, * g;
};

string gcontent;
unsigned diam = 8, dx = 10, dy = 12;
const double mmtoin = 1/25.4;
const double mmtopt = 72*mmtoin;
unsigned lineno;

const char * errmsg0 = "Error: incorrect program invocation.\n";
const char * errmsg1 = "Error: incorrect heading.\n";
const char * errmsg2 = "Error: tree empty.\n";
const char * errmsg3 = "Error: incorrect content at line %d.\n";
const char * errmsg4 = "Error: tree ended prematurely at line %d.\n";
const char * errmsg5 = "Error: incorrect offset at line %d.\n";
const char * errmsg6 = "Error: line %d is non-empty but appears to be not in the tree.\n";

line * getln() {
  auto x = new line;
  if (!getline(cin,x->str))
    x->off = -1;
  else {
    auto i = x->str.find_first_not_of(" ");
    if (i == string::npos)
      x->off = -1;
    else {
      x->off = i;
      x->str = x->str.substr(i,x->str.find_last_not_of(" ")+1-i);
    }
    ++lineno;
  }
  return x;
}

bool settings(const string & s) {
  regex rx("([^[]*)(?:\\[([^\\]]+)\\])?");
  smatch rs;
  if (!regex_match(s,rs,rx) || (rs[1].str().empty() && rs[2].str().empty())) return false;
  gcontent = rs[2].str();
  istringstream sstr(rs[1].str());
  string w;
  size_t n;
  while (sstr >> w)
    if (w == "diam") {
      if (!(sstr >> w) || (diam = stoul(w,&n), n != w.size())) return false;
    } else if (w == "dx") {
      if (!(sstr >> w) || (dx = stoul(w,&n), n != w.size())) return false;
    } else if (w == "dy") {
      if (!(sstr >> w) || (dy = stoul(w,&n), n != w.size())) return false;
    }
    else return false;
  return true;
}

void perrt(const char * s, int n) {
  fprintf(stderr,s,n);
  throw 0;
}

void parseln(line * x, string * & sc, string * & sn, string * & se) {
  static const string cesc("-]\\");
  static const string pesc("\\^$.+*?|()[]{}");
  string c(x->str.substr(0,1));
  string s1 = cesc.find(c[0]) == string::npos ? c : "\\"+c;
  string s2 = pesc.find(c[0]) == string::npos ? c : "\\"+c;
  regex rx(s2+"([^"+s1+"]*)"+s2+"\\s*(?:\\[([^\\]]+)\\])?(?:\\s*\\{([^}]+)\\})?");
  smatch rs;
  if (!regex_match(x->str,rs,rx))
    perrt(errmsg3,lineno);
  sc = new string(rs[1].str());
  string s;
  s = rs[2].str();
  if (!s.empty()) sn = new string(s);
  s = rs[3].str();
  if (!s.empty()) se = new string(s);
}

line * rtbuild(RTnode * & t, line * s) {
  while (s->str.size() == 1) {   // skip dummy nodes
    s = getln();
    if (s->off == -1)
      perrt(errmsg4,lineno);
  }
  string * sc = nullptr, * sn = nullptr, * se = nullptr;
  parseln(s,sc,sn,se);
  t = new RTnode{lineno,sc,sn,se,nullptr};
  line * p, * q;
  do p = getln(); while (p->off > -1 && p->str.size() == 1);
  if (p->off <= s->off)       // no branches
    return p;
  t->desc = new vector<RTnode *>(1);
  for (int i=0;; ++i) {
    q = rtbuild((*(t->desc))[i],p);
    if (q->off != p->off) {   // no more siblings or no more nodes at all
      if (q->off > p->off)
        perrt(errmsg5,lineno);
      return q;
    }
    t->desc->resize(1+t->desc->size());
    p = q;
  }
}

line * btbuild(BTnode * & t, line * s) {
  string * sc = nullptr, * sn = nullptr, * se = nullptr;
  parseln(s,sc,sn,se);
  t = new BTnode{lineno,sc,sn,se,nullptr,nullptr};
  line * f = getln();
  if (f->off <= s->off)       // no branches
    return f;
  line * g;
  if (f->str.size() == 1) {   // empty left branch
    g = getln();
    if (g->off == -1)
      perrt(errmsg4,lineno);
  } else {                    // non-empty left branch
    g = btbuild(t->f,f);
    if (g->off < f->off)      // no right branch or no more nodes at all
      return g;
  }
  if (g->off != f->off)
    perrt(errmsg5,lineno);
  return btbuild(t->g,g);
}

void outputnode(unsigned n, string * sc, string * sn, double x, double y) {
  unsigned kx = .5+x*dx*mmtopt, ky = .5+y*dy*mmtopt;
  if (xychange) swap(kx,ky);
  cout << "  \"" << n << "\" [label=\"" << *sc << "\" pos=\"" << kx << ',' << ky << '"';
  if (sn != nullptr)
    cout << ' ' << *sn;
  cout << "];\n";
}

void rtwalk(RTnode * t) {
  double x = 0, y = 0;
  function<void(RTnode *)> fw = [&](RTnode * t) {
    if (t == nullptr) return;
    unsigned nd = t->desc == nullptr ? 0 : t->desc->size();
    double x0 = x;
    if (nd > 0) {
      for (auto n : *(t->desc)) {         // output edges or arcs from t to its descendants
        cout << "  \"" << t->label << "\" -" << (digraph ? '>' : '-') << '"' << n->label << '"';
        if (n->attedge != nullptr)
          cout << " [" << *(n->attedge) << ']';
        cout << ";\n";
      }
      ++y;
      for (unsigned i=0; i < nd; ++i)
        fw((*(t->desc))[i]);
      --y;
    }
    outputnode(t->label,t->content,t->attnode,nd>0?(x0+x-1)/2:x,y);
    if (nd == 0) ++x;
  };
  if (digraph) cout << "di";
  cout << "graph {\n  node [shape=circle width=" << diam*mmtoin << " fixedsize=true fontname=\"sans-serif\" fontsize=11";
  if (!gcontent.empty())
    cout << ' ' << gcontent;
  cout << "];\n";
  fw(t);
  cout << "}\n";
}

void btwalk(BTnode * t) {
  double x = 0, y = 0;
  function<void(BTnode *)> fw = [&](BTnode * t) {
    auto fh = [&t](BTnode * node) {    // output an edge or arc from t to node
      if (node != nullptr) {
        cout << "  \"" << t->label << "\" -" << (digraph ? '>' : '-') << '"' << node->label << '"';
        if (node->attedge != nullptr)
          cout << " [" << *(node->attedge) << ']';
        cout << ";\n";
      }
    };
    if (t == nullptr) return;
    fh(t->f); fh(t->g);
    ++y;
    fw(t->f);
    outputnode(t->label,t->content,t->attnode,x,y-1);
    ++x;
    fw(t->g);
    --y;
  };
  if (digraph) cout << "di";
  cout << "graph {\n  node [shape=circle width=" << diam*mmtoin << " fixedsize=true fontname=\"sans-serif\" fontsize=11";
  if (!gcontent.empty())
    cout << ' ' << gcontent;
  cout << "];\n";
  fw(t);
  cout << "}\n";
}

bool params(char * s) {
  unsigned b = 0, d = 0, x = 0;
  for (unsigned i = 0; *s; ++s, ++i)
    if (i == 3) return false;
    else if (*s == 'b') {binary = true; ++b;}
    else if (*s == 'd') {digraph = true; ++d;}
    else if (*s == 'x') {xychange = true; ++x;}
    else return false;
  return b <= 1 && d <=1 && x <= 1;
}

int main(int argc, char * argv[]) {
  binary = digraph = xychange = false;
  try {
    if (argc > 2 || (argc == 2 && !params(argv[1])))
      perrt(errmsg0,0);
    lineno = 0;
    auto n = getln();
    if (n->off == 0 && n->str[0] == '=') {
      if (settings(n->str.substr(1))) {
        n = getln();
      } else
        perrt(errmsg1,0);
    }
    if (n->off == -1)
      perrt(errmsg2,0);
    if (binary) {
      BTnode * t;
      if (btbuild(t,n)->off == -1)
        btwalk(t);
      else
        perrt(errmsg6,lineno);
    } else {
      RTnode * t;
      if (rtbuild(t,n)->off == -1)
        rtwalk(t);
      else
        perrt(errmsg5,lineno);
    }
  } catch(...) {
    return -1;
  }
}
